#!/usr/bin/env python3

import json
import subprocess
from pathlib import Path
from types import TracebackType
from typing import Any, Dict, List, NamedTuple, Optional, cast


# We use NamedTuple instead of dataclasses for Python3.5/6 support.
class PyreCheckResult(NamedTuple):
    exit_code: int
    errors: Optional[List[str]]


PyreQueryResult = Dict[str, Any]


class PyreConnection:
    def __init__(self, pyre_directory: Optional[Path] = None) -> None:
        self.pyre_directory: Path = (
            pyre_directory if pyre_directory is not None else Path.cwd()
        )
        self.server_initialized = False

    def __enter__(self) -> "PyreConnection":
        self.start_server()
        return self

    def __exit__(
        self,
        _type: Optional[BaseException],
        _value: Optional[BaseException],
        _traceback: Optional[TracebackType],
    ) -> None:
        self.stop_server()
        return None

    def start_server(self) -> PyreCheckResult:
        subprocess.run(["pyre", "start"], cwd=str(self.pyre_directory))
        result = subprocess.run(
            ["pyre", "incremental"],
            stdout=subprocess.PIPE,
            cwd=str(self.pyre_directory),
        )
        self.server_initialized = True
        return _parse_check_output(result)

    def restart_server(self) -> PyreCheckResult:
        result = _parse_check_output(
            subprocess.run(
                ["pyre", "restart"],
                stdout=subprocess.PIPE,
                cwd=str(self.pyre_directory),
            )
        )
        self.server_initialized = True
        return result

    def stop_server(self) -> None:
        subprocess.run(["pyre", "stop"], check=True, cwd=str(self.pyre_directory))

    def query_server(self, query: str) -> Optional[PyreQueryResult]:
        if not self.server_initialized:
            self.start_server()
        result = subprocess.run(
            ["pyre", "query", query],
            stdout=subprocess.PIPE,
            cwd=str(self.pyre_directory),
        )
        if result.returncode != 0:
            return None
        return json.loads(result.stdout.decode())


def _parse_check_output(
    completed_process: "subprocess.CompletedProcess[bytes]"
) -> PyreCheckResult:
    errors = completed_process.stdout.decode().split()
    return PyreCheckResult(exit_code=completed_process.returncode, errors=errors)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import functools
import logging
import os
import shutil
import subprocess
from itertools import chain
from pathlib import Path
from time import time
from typing import Dict, List, NamedTuple, Optional, Set, Tuple

from . import _resolve_filter_paths, buck, filesystem, json_rpc, log
from .configuration import Configuration
from .exceptions import EnvironmentException
from .filesystem import (
    BuckBuilder,
    _compute_symbolic_link_mapping,
    _delete_symbolic_link,
    acquire_lock,
    add_symbolic_link,
    find_python_paths,
    is_empty,
    is_parent,
    remove_if_exists,
    translate_paths,
)
from .socket_connection import SocketConnection, SocketException


LOG: logging.Logger = logging.getLogger(__name__)


# If there are a lot of tracked files that are updated at the same time, it is
# probably a rebase. So, rebuild just to be safe.
REBUILD_THRESHOLD_FOR_UPDATED_PATHS: int = 4


# If there are a lot of new or deleted files, it is probably a rebase.
# This is a separate threshold from the number of updated tracked files because
# we don't know for sure if these new files should actually be tracked. So, we
# might want a higher threshold in order to avoid rebuilding for spurious new
# files.
REBUILD_THRESHOLD_FOR_NEW_OR_DELETED_PATHS: int = 5


DONT_CARE_PROGRESS_VALUE = 1


class UpdatedPaths(NamedTuple):
    updated_paths: List[str]
    deleted_paths: List[str]

    def is_empty(self) -> bool:
        return not self.updated_paths and not self.deleted_paths


class AnalysisDirectory:
    def __init__(
        self,
        path: str,
        filter_paths: Optional[List[str]] = None,
        search_path: Optional[List[str]] = None,
    ) -> None:
        self._path = path
        self._filter_paths: List[str] = filter_paths or []
        self._search_path: List[str] = search_path or []

    def get_root(self) -> str:
        return self._path

    def get_filter_root(self) -> List[str]:
        return self._filter_paths or [self.get_root()]

    def prepare(self) -> None:
        pass

    def compute_symbolic_links(self) -> Dict[str, str]:
        return {}

    def process_updated_files(self, paths: List[str]) -> UpdatedPaths:
        """
            Process a list of paths which were added/removed/updated, making any
            necessary changes to the directory:
                - For an AnalysisDirectory, nothing needs to be changed, since
                  the mapping from source file to analysis file is 1:1.
                - For a SharedAnalysisDirectory, the symbolic links (as well as
                  the reverse-mapping we track) need to be updated to account for
                  new and deleted files.

            Return a list of files (corresponding to the given paths) that Pyre
            should be tracking.
        """
        deleted_paths = [path for path in paths if not os.path.isfile(path)]
        tracked_paths = [
            path
            for path in paths
            if self._is_tracked(path) and path not in deleted_paths
        ]
        tracked_paths.extend(deleted_paths)
        return UpdatedPaths(updated_paths=tracked_paths, deleted_paths=deleted_paths)

    def cleanup(self) -> None:
        pass

    @property
    @functools.lru_cache(1)
    def _tracked_directories(self) -> List[str]:
        tracked_directories = [
            self.get_root(),
            *[os.path.join(*path.split("$")) for path in self._search_path],
        ]
        return [os.path.abspath(path) for path in tracked_directories]

    def _is_tracked(self, path: str) -> bool:
        return any(
            is_parent(directory, path) for directory in self._tracked_directories
        )


class SharedAnalysisDirectory(AnalysisDirectory):
    def __init__(
        self,
        source_directories: List[str],
        targets: List[str],
        original_directory: Optional[str] = None,
        filter_paths: Optional[List[str]] = None,
        local_configuration_root: Optional[str] = None,
        extensions: Optional[List[str]] = None,
        search_path: Optional[List[str]] = None,
        isolate: bool = False,
        buck_builder: Optional[BuckBuilder] = None,
        configuration: Optional[Configuration] = None,
    ) -> None:
        self._source_directories: Set[str] = set(source_directories)
        self._targets: Set[str] = set(targets)
        self._original_directory = original_directory
        self._filter_paths: List[str] = filter_paths or []
        self._local_configuration_root = local_configuration_root
        self._extensions: Set[str] = set(extensions or []) | {"py", "pyi"}
        self._search_path: List[str] = search_path or []
        self._isolate = isolate
        self._buck_builder: BuckBuilder = buck_builder or buck.SimpleBuckBuilder()

        # Mapping from source files in the project root to symbolic links in the
        # analysis directory.
        self._symbolic_links = {}  # type: Dict[str, str]

        self._configuration = configuration

    def get_scratch_directory(self) -> str:
        try:
            return (
                subprocess.check_output(["scratch", "path", "--subdir", "pyre"])
                .decode("utf-8")
                .strip()
            )
        except Exception:
            return os.path.join(os.getcwd(), ".pyre")

    @functools.lru_cache(1)
    def get_root(self) -> str:
        path_to_root = self._local_configuration_root or "shared_analysis_directory"
        suffix = "_{}".format(str(os.getpid())) if self._isolate else ""
        return os.path.join(
            self.get_scratch_directory(), "{}{}".format(path_to_root, suffix)
        )

    def get_filter_root(self) -> List[str]:
        return self._filter_paths or [os.getcwd()]

    # Exposed for testing.
    def _resolve_source_directories(self) -> None:
        if self._targets:
            new_source_directories = self._buck_builder.build(self._targets)
            original_directory = self._original_directory
            if original_directory is not None:
                new_source_directories = translate_paths(
                    # pyre-fixme[6]: Expected `Set[str]` for 1st anonymous parameter
                    # to call `translate_paths` but got `typing.Iterable[str]`.
                    new_source_directories,
                    original_directory,
                )
            self._source_directories.update(new_source_directories)

        if len(self._source_directories) == 0:
            raise EnvironmentException("No targets or source directories to analyze.")

    def prepare(self) -> None:
        start = time()
        root = self.get_root()
        LOG.info("Constructing shared directory `%s`", root)

        self._resolve_source_directories()

        try:
            os.makedirs(root)
        except OSError:
            pass  # Swallow.

        lock = os.path.join(root, ".pyre.lock")
        with acquire_lock(lock, blocking=True):
            self._clear()
            self._merge()
            LOG.log(
                log.PERFORMANCE, "Merged analysis directories in %fs", time() - start
            )
        self._symbolic_links.update(self.compute_symbolic_links())

    def rebuild(self) -> None:
        start = time()

        root = self.get_root()
        LOG.info("Updating shared directory `%s`", root)

        self._resolve_source_directories()

        with filesystem.acquire_lock(os.path.join(root, ".pyre.lock"), blocking=True):
            all_paths = {}
            for source_directory in self._source_directories:
                self._merge_into_paths(source_directory, all_paths)
            for relative_path, project_path in all_paths.items():
                scratch_path = os.path.join(root, relative_path)
                if os.path.realpath(scratch_path) != project_path:
                    add_symbolic_link(scratch_path, project_path)
            for scratch_path in self._symbolic_links.values():
                if not os.path.exists(scratch_path):
                    os.remove(scratch_path)
            LOG.log(log.PERFORMANCE, "Updated shared directory in %fs", time() - start)
        self._symbolic_links = self.compute_symbolic_links()

    def compute_symbolic_links(self) -> Dict[str, str]:
        return _compute_symbolic_link_mapping(self.get_root(), self._extensions)

    @staticmethod
    def should_rebuild(
        updated_tracked_paths: List[str], new_paths: List[str], deleted_paths: List[str]
    ) -> bool:
        if any(
            Path(path).suffix == ".thrift" or Path(path).name == "TARGETS"
            for path in chain(updated_tracked_paths, new_paths, deleted_paths)
        ):
            return True
        return (
            len(updated_tracked_paths) >= REBUILD_THRESHOLD_FOR_UPDATED_PATHS
            or len(new_paths) + len(deleted_paths)
            >= REBUILD_THRESHOLD_FOR_NEW_OR_DELETED_PATHS
        )

    def _notify_about_rebuild(self, is_start_message: bool = True) -> None:
        configuration = self._configuration
        if configuration is None:
            return

        LOG.info(
            "Notifying server of %s of rebuild.", "start" if is_start_message else "end"
        )
        if is_start_message:
            message = (
                "Pyre is rebuilding because a significant number of files were "
                "changed. Your results may be outdated until this is finished."
            )
            short_message = "Rebuilding..."
            message_type = json_rpc.LanguageServerMessageType.WARNING.value
        else:
            message = "Done rebuilding."
            short_message = "Done rebuilding."
            message_type = json_rpc.LanguageServerMessageType.INFORMATION.value
        show_status_message = json_rpc.Request(
            method="window/showStatus",
            parameters={
                "message": message,
                "shortMessage": short_message,
                "type": message_type,
                "actions": [],
                "progress": {
                    "numerator": DONT_CARE_PROGRESS_VALUE,
                    "denominator": DONT_CARE_PROGRESS_VALUE,
                },
            },
        )
        try:
            with SocketConnection(configuration.log_directory) as socket_connection:
                socket_connection.perform_handshake(configuration.version_hash)
                socket_connection.send_request(show_status_message)
        except (
            SocketException,
            ResourceWarning,
            json_rpc.JSONRPCException,
        ) as exception:
            LOG.error("Error while communicating with server: %s", str(exception))

    def _process_rebuilt_files(
        self, tracked_paths: List[str], deleted_paths: List[str]
    ) -> Tuple[List[str], List[str]]:
        self._notify_about_rebuild(is_start_message=True)

        old_scratch_paths = set(self._symbolic_links.keys())
        self.rebuild()
        new_scratch_paths = set(self._symbolic_links.keys())

        self._notify_about_rebuild(is_start_message=False)

        # We ignore the individual new_paths from above and consider only
        # paths updated during a rebuild.
        tracked_paths.extend(new_scratch_paths - old_scratch_paths)
        deleted_paths = list(old_scratch_paths - new_scratch_paths)
        return tracked_paths, deleted_paths

    def _process_new_paths(
        self, new_paths: List[str], tracked_paths: List[str]
    ) -> List[str]:
        relative_link_map = {}
        try:
            relative_link_map = buck.query_buck_relative_paths(new_paths, self._targets)
        except buck.BuckException as error:
            LOG.error("Exception occurred when querying buck: %s", error)
            LOG.error("No new paths will be added to the analysis directory.")

        absolute_link_map = {
            path: os.path.join(self.get_root(), relative_link)
            for path, relative_link in relative_link_map.items()
        }
        tracked_paths.extend(absolute_link_map.keys())
        for path, absolute_link in absolute_link_map.items():
            try:
                add_symbolic_link(absolute_link, path)
                self._symbolic_links[path] = absolute_link
            except OSError:
                LOG.warning("Failed to add link at %s.", absolute_link)
        return tracked_paths

    def _process_deleted_paths(self, deleted_paths: List[str]) -> List[str]:
        deleted_links = [
            project_path
            for project_path in deleted_paths
            if project_path in self._symbolic_links
        ]
        for path in deleted_paths:
            link = self._symbolic_links.pop(path, None)
            if link:
                try:
                    _delete_symbolic_link(link)
                except OSError:
                    LOG.warning("Failed to delete link at `%s`.", link)
        return deleted_links

    def process_updated_files(self, paths: List[str]) -> UpdatedPaths:
        """Update the analysis directory for any new or deleted files.
        Rebuild the directory using buck if needed.
        Return the updated and deleted paths."""
        tracked_paths = []
        deleted_paths = [path for path in paths if not os.path.isfile(path)]
        new_paths = [
            path
            for path in paths
            if path not in self._symbolic_links
            and os.path.isfile(path)
            and is_parent(os.getcwd(), path)
        ]
        updated_paths = [
            path
            for path in paths
            if path not in deleted_paths and path not in new_paths
        ]

        for path in updated_paths:
            if path in self._symbolic_links or self._is_tracked(path):
                tracked_paths.append(path)

        if SharedAnalysisDirectory.should_rebuild(
            tracked_paths, new_paths, deleted_paths
        ):
            tracked_paths, deleted_paths = self._process_rebuilt_files(
                tracked_paths, deleted_paths
            )
        elif new_paths or deleted_paths:
            if new_paths:
                LOG.info("Detected new paths: %s.", ",".join(new_paths))
                tracked_paths = self._process_new_paths(new_paths, tracked_paths)
            if deleted_paths:
                LOG.info("Detected deleted paths: `%s`.", "`,`".join(deleted_paths))
                deleted_paths = self._process_deleted_paths(deleted_paths)

        tracked_paths.extend(deleted_paths)
        return UpdatedPaths(updated_paths=tracked_paths, deleted_paths=deleted_paths)

    def cleanup(self) -> None:
        try:
            if self._isolate:
                shutil.rmtree(self.get_root())
        except Exception:
            pass

    def _clear(self) -> None:
        root = self.get_root()
        for path in os.listdir(root):
            if path.startswith(".pyre"):
                continue

            path = os.path.join(root, path)
            remove_if_exists(path)

    def _merge(self) -> None:
        root = self.get_root()

        all_paths = {}
        for source_directory in self._source_directories:
            self._merge_into_paths(source_directory, all_paths)
        for relative, original in all_paths.items():
            merged = os.path.join(root, relative)
            add_symbolic_link(merged, original)

    # Exposed for testing.
    def _merge_into_paths(
        self, source_directory: str, all_paths: Dict[str, str]
    ) -> None:
        paths = find_python_paths(root=source_directory)
        for path in paths:
            relative = os.path.relpath(path, source_directory)
            if not path:
                continue
            # don't bother stat'ing paths that are already in the analysis directory.
            if relative in all_paths:
                continue
            try:
                absolute = os.path.realpath(path)
                # Don't merge symlinked directories.
                if not os.path.isfile(absolute):
                    continue
                if relative.endswith("__init__.py") and is_empty(absolute):
                    # Don't let empty __init__.py files override legitimate files.
                    continue
                all_paths[relative] = absolute
            except FileNotFoundError:
                continue


def resolve_analysis_directory(
    arguments: argparse.Namespace,
    configuration: Configuration,
    original_directory: str,
    current_directory: str,
    build: bool = False,
    isolate: bool = False,
) -> AnalysisDirectory:
    # Only read from the configuration if no explicit targets are passed in.
    if not arguments.source_directories and not arguments.targets:
        source_directories = configuration.source_directories
        targets = configuration.targets
    else:
        source_directories = arguments.source_directories or []
        targets = arguments.targets or []
        if targets:
            configuration_name = ".pyre_configuration.local"
            command = "pyre init --local"
        else:
            configuration_name = ".pyre_configuration"
            command = "pyre init"
        LOG.warning(
            "Setting up a `%s` with `%s` may reduce overhead.",
            configuration_name,
            command,
        )

    if arguments.filter_directory:
        filter_paths = [arguments.filter_directory]
    else:
        filter_paths = _resolve_filter_paths(
            arguments, configuration, original_directory
        )

    local_configuration_root = configuration.local_configuration_root
    if local_configuration_root:
        local_configuration_root = os.path.relpath(
            local_configuration_root, current_directory
        )

    use_buck_builder = (
        arguments.use_buck_builder or configuration.use_buck_builder
    ) and not arguments.use_legacy_builder

    if len(source_directories) == 1 and len(targets) == 0:
        analysis_directory = AnalysisDirectory(
            source_directories.pop(),
            filter_paths=filter_paths,
            search_path=configuration.search_path,
        )
    else:
        build = arguments.build or build
        buck_builder = buck.SimpleBuckBuilder(build=build)
        if use_buck_builder:
            buck_root = buck.find_buck_root(os.getcwd())
            if not buck_root:
                raise EnvironmentException(
                    "No Buck configuration at `{}` or any of its ancestors.".format(
                        os.getcwd()
                    )
                )
            buck_builder = buck.FastBuckBuilder(
                buck_root=buck_root,
                buck_builder_binary=arguments.buck_builder_binary,
                buck_builder_target=arguments.buck_builder_target,
                debug_mode=arguments.buck_builder_debug,
            )
        else:
            buck_builder = buck.SimpleBuckBuilder(build=build)

        analysis_directory = SharedAnalysisDirectory(
            source_directories=source_directories,
            targets=targets,
            buck_builder=buck_builder,
            original_directory=original_directory,
            filter_paths=filter_paths,
            local_configuration_root=local_configuration_root,
            extensions=configuration.extensions,
            search_path=configuration.search_path,
            isolate=isolate,
            configuration=configuration,
        )
    return analysis_directory



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import logging
import os
import time
from logging import Logger
from typing import List, Optional

from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from ..project_files_monitor import ProjectFilesMonitor
from .command import ClientException, Command, State
from .kill import Kill


LOG: Logger = logging.getLogger(__name__)


class Stop(Command):
    NAME = "stop"

    def __init__(
        self,
        arguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Stop, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        stop = parser.add_parser(cls.NAME, epilog="Signals the Pyre server to stop.")
        stop.set_defaults(command=cls)

    def _flags(self) -> List[str]:
        log_directory = self._log_directory
        flags = []
        if log_directory:
            flags.extend(["-log-directory", log_directory])
        return flags

    def _run(self) -> None:
        def _kill():
            arguments = self._arguments
            arguments.with_fire = False
            Kill(
                arguments,
                self._original_directory,
                self._configuration,
                self._analysis_directory,
            ).run()

        if self._state() == State.DEAD:
            LOG.warning("No server running, cleaning up any left over Pyre processes.")
            _kill()
        else:
            try:
                with open(
                    os.path.join(self._log_directory, "server", "server.pid")
                ) as pid_file:
                    pid_to_poll = int(pid_file.read())
            except (OSError, ValueError):
                pid_to_poll = None
            try:
                stopped = False
                # If this call fails, check() will throw a ClientException.
                self._call_client(command=self.NAME).check()
                # Poll for a second to ensure that the server has a chance to exit.
                if pid_to_poll is not None:
                    stop_time = time.time() + 1.0
                    LOG.info("Polling for server's process to stop...")
                    while time.time() < stop_time:
                        # send a null signal to validate the process's existence. If the
                        # process has terminated, a ProcessLookupError will be thrown.
                        os.kill(pid_to_poll, 0)
                        time.sleep(0.1)
                stopped = True
            except ClientException:
                # An error was encountered when running `pyre stop`.
                stopped = False
            except ProcessLookupError:
                LOG.info("The server process has stopped.")
                stopped = True

            if not stopped:
                LOG.warning("Could not stop server, attempting to kill.")
                _kill()
            else:
                LOG.info("Stopped server at `%s`", self._analysis_directory.get_root())

        ProjectFilesMonitor.stop_project_monitor(self._configuration)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import logging
from logging import Logger
from typing import List, Optional

from ..analysis_directory import AnalysisDirectory, resolve_analysis_directory
from ..configuration import Configuration
from .command import ExitCode, typeshed_search_path
from .reporting import Reporting


LOG: Logger = logging.getLogger(__name__)


class Check(Reporting):
    NAME = "check"

    def __init__(
        self,
        arguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Check, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        check = parser.add_parser(
            cls.NAME,
            epilog="""
          Runs a one-time check of a project without initializing a type check server.
        """,
        )
        check.set_defaults(command=cls)

    def generate_analysis_directory(self) -> AnalysisDirectory:
        return resolve_analysis_directory(
            self._arguments,
            self._configuration,
            self._original_directory,
            self._current_directory,
            build=True,
            isolate=True,
        )

    def _flags(self) -> List[str]:
        flags = super()._flags()
        filter_directories = self._get_directories_to_analyze()
        if len(filter_directories):
            flags.extend(["-filter-directories", ";".join(sorted(filter_directories))])
        flags.extend(["-workers", str(self._number_of_workers)])
        search_path = self._configuration.search_path + typeshed_search_path(
            self._configuration.typeshed
        )
        if len(self._configuration.ignore_all_errors):
            flags.extend(
                [
                    "-ignore-all-errors",
                    ";".join(sorted(self._configuration.ignore_all_errors)),
                ]
            )
        if search_path:
            flags.extend(["-search-path", ",".join(search_path)])
        excludes = self._configuration.excludes
        for exclude in excludes:
            flags.extend(["-exclude", exclude])
        extensions = self._configuration.extensions
        for extension in extensions:
            flags.extend(["-extension", extension])
        return flags

    def _run(self, retries: int = 1) -> None:
        self._analysis_directory.prepare()

        result = self._call_client(command="check")
        errors = self._get_errors(result)
        self._print(errors)

        if errors:
            self._exit_code = ExitCode.FOUND_ERRORS



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe
import argparse
import json
from enum import Enum
from typing import List, Optional

from .. import log
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from .command import Command


# this needs a better name
class TypeAnnotation:
    def __init__(
        self,
        coverage: str,
        start_line: int,
        stop_line: int,
        start_column: int,
        stop_column: int,
    ) -> None:
        self.coverage = coverage
        self.start_line = start_line
        self.stop_line = stop_line
        self.start_column = start_column
        self.stop_column = stop_column

    def __hash__(self) -> int:
        return hash(
            (self.start_line, self.stop_line, self.start_column, self.stop_column)
        )

    def __eq__(self, other: object):
        if not isinstance(other, TypeAnnotation):
            return False
        return (
            self.start_line == other.start_line
            and self.stop_line == other.stop_line
            and self.start_column == other.start_column
            and self.stop_column == other.stop_column
        )

    @staticmethod
    def create_from_json(data) -> "TypeAnnotation":
        start = data["location"]["start"]
        stop = data["location"]["stop"]
        type = data["coverage"][0]

        if type == CoverageLevel.TYPED.value:
            coverage = CoverageLevel.TYPED
        elif type == CoverageLevel.PARTIAL.value:
            coverage = CoverageLevel.PARTIAL
        else:
            coverage = CoverageLevel.UNTYPED

        return TypeAnnotation(
            # pyre-fixme[6]: Expected `str` for 1st param but got `CoverageLevel`.
            coverage,
            start["line"],
            stop["line"],
            start["column"],
            stop["column"],
        )


class CoverageLevel(Enum):
    TYPED = "Typed"
    PARTIAL = "Partial"
    UNTYPED = "Untyped"


class PrintColor:
    def __init__(self, type_annotations, path) -> None:
        self.type_annotations = type_annotations  # List[TypeAnnotation]
        self.path = path

    def _add_color(self, line: str, type: CoverageLevel) -> str:
        coverage_to_color = {
            CoverageLevel.TYPED: log.Color.GREEN,
            CoverageLevel.PARTIAL: log.Color.YELLOW,
            CoverageLevel.UNTYPED: log.Color.RED,
        }
        return coverage_to_color[type] + line + log.Format.CLEAR

    def _find_types_at_line(self, line_number: int):  # List
        types_at_line = []
        for type in self.type_annotations:
            if type.start_line == line_number:
                types_at_line.append(type)
        return types_at_line

    def print_results(self) -> None:
        with open(self.path) as file:
            lines = file.read().split("\n")
            for line_number, line in enumerate(lines, start=1):
                type_annotations = self._find_types_at_line(line_number)
                type_annotations.sort(key=lambda annotation: annotation.start_column)
                for annotation in reversed(type_annotations):
                    prefix = line[: annotation.start_column]
                    string = line[annotation.start_column : annotation.stop_column]
                    suffix = line[annotation.stop_column :]
                    string_with_color = self._add_color(string, annotation.coverage)
                    line = prefix + string_with_color + suffix
                log.stdout.write(line + "\n")


class Color(Command):
    NAME = "color"

    def __init__(
        self,
        arguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Color, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self.path = arguments.path

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        color = parser.add_parser(cls.NAME)
        color.set_defaults(command=cls)
        color.add_argument("path")

    def _flags(self) -> List[str]:
        return ["coverage_in_file('" + self.path + "')"]

    def _run(self) -> None:
        result = self._call_client(command="query")
        result.check()
        if "error" in result.output:
            log.stdout.write(result.output)
            return

        # convert result to TypeAnnotation
        coverage = json.loads(result.output)
        type_annotations = [
            TypeAnnotation.create_from_json(type)
            for type in coverage["response"]["types"]
        ]
        PrintColor(set(type_annotations), self.path).print_results()



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import logging
import os
from logging import Logger
from typing import Any, Dict, List

from .analysis_directory import AnalysisDirectory
from .commands import stop
from .configuration import CONFIGURATION_FILE
from .filesystem import is_parent
from .watchman_subscriber import Subscription, WatchmanSubscriber


LOG: Logger = logging.getLogger(__name__)


class ConfigurationMonitor(WatchmanSubscriber):
    """
        The ConfigurationMonitor watches only for .pyre_configuration(.local)
        files, and will kill the corresponding server and ProjectFileMonitor
        when a configuration changes.
        Logs are found in .pyre/configuration_monitor/configuration_monitor.log
        To kill a  monitor, get pid from
        .pyre/configuration_monitor/configuration_monitor.pid ; kill <pid>.
    """

    def __init__(
        self,
        arguments,
        configuration,
        analysis_directory: AnalysisDirectory,
        project_root: str,
    ) -> None:
        base_path = os.path.join(project_root, ".pyre")
        super(ConfigurationMonitor, self).__init__(base_path)
        self.arguments = arguments
        self.configuration = configuration
        self.analysis_directory = analysis_directory
        self.project_root = project_root

    @property
    def _name(self) -> str:
        return "configuration_monitor"

    @property
    def _subscriptions(self) -> List[Subscription]:
        roots = self._watchman_client.query("watch-list")["roots"]
        for root in roots:
            if os.path.commonprefix(
                [os.path.realpath(root), os.path.realpath(self.project_root)]
            ):
                name = "pyre_monitor_{}".format(os.path.basename(root))
                subscription = {
                    "empty_on_fresh_instance": True,
                    "expression": [
                        "allof",
                        ["type", "f"],
                        ["not", "empty"],
                        [
                            "anyof",
                            ["suffix", "pyre_configuration.local"],
                            ["suffix", "pyre_configuration"],
                        ],
                    ],
                    "fields": ["name"],
                }
                return [Subscription(root, name, subscription)]
        return []

    def _handle_response(self, response: Dict[str, Any]) -> None:
        LOG.info(
            "Update to configuration at %s",
            os.path.join(response["root"], ",".join(response["files"])),
        )
        absolute_path = [
            os.path.join(response["root"], file) for file in response["files"]
        ]

        # Find the path to the project configuration file and compare it with the
        # list of changed configuration files.
        project_configuration = os.path.join(response["root"], CONFIGURATION_FILE)
        if any((project_configuration == file) for file in absolute_path):
            LOG.info("Pyre configuration changed. Stopping pyre server.")
            stop.Stop(self.arguments, self.configuration, self.analysis_directory).run()
            # TODO(T54088045): Find all local pyre servers running underneath
            # and stop them.
        else:
            LOG.info("None of the changed paths correspond to pyre configuration.")

        if self.arguments.local_configuration:
            if any(
                is_parent(self.arguments.local_configuration, file)
                for file in absolute_path
            ):
                LOG.info("Local configuration changed. Stopping pyre server.")
                stop.Stop(
                    self.arguments, self.configuration, self.analysis_directory
                ).run()
            else:
                LOG.info(
                    "None of the changed paths correspond to the current local"
                    "configuration."
                )



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import ast
import logging
import pathlib
from logging import Logger
from typing import Callable

from pyre_extensions import ListVariadic


Ts = ListVariadic("Ts")


LOG: Logger = logging.getLogger(__name__)


def verify_stable_ast(file_modifier: Callable[[Ts], None]) -> Callable[[Ts], None]:
    def wrapper(arguments: argparse.Namespace, filename: str, *args):
        # AST before changes
        path = pathlib.Path(filename)
        try:
            text = path.read_text()
            ast_before = ast.parse(text)

            # AST after changes
            file_modifier(arguments, filename, *args)
            new_text = path.read_text()
            try:
                ast_after = ast.parse(new_text)

                # Undo changes if AST does not match
                if not ast.dump(ast_before) == ast.dump(ast_after):
                    LOG.warning(
                        "Attempted file changes modified the AST in %s. Undoing.",
                        filename,
                    )
                    path.write_text(text)
            except Exception as e:
                LOG.warning("Could not parse file %s. Undoing.", filename)
                LOG.warning(e)
                path.write_text(text)
        except FileNotFoundError:
            LOG.warning("File %s cannot be found, skipping.", filename)
            return

    return wrapper



# Copyright (c) 2019-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import json
from enum import Enum
from json.decoder import JSONDecodeError
from typing import Any, BinaryIO, Dict, Optional


JSON = Dict[str, Any]


class LanguageServerMessageType(Enum):
    """Message type for an LSP warning message."""

    WARNING = 2
    INFORMATION = 3


class JSONRPCException(Exception):
    pass


class Request(object):
    def __init__(
        self, method: str, id: Optional[str] = None, parameters: Optional[JSON] = None
    ) -> None:
        self.method = method
        self.id = id
        self.parameters = parameters

    def json(self) -> str:
        object = {"jsonrpc": "2.0", "method": self.method}  # type: Dict[str, Any]
        if self.id:
            object["id"] = self.id
        if self.parameters:
            object["params"] = self.parameters
        return json.dumps(object)

    @staticmethod
    def validate_payload(payload: JSON) -> bool:
        return payload.get("jsonrpc") == "2.0" and payload.get("method") is not None

    def write(self, file: BinaryIO) -> bool:
        try:
            payload = self.json()
            length = len(payload.encode("utf-8"))

            response = ("Content-Length: {}\r\n\r\n{}").format(length, payload)
            file.write(response.encode("utf-8"))
            file.flush()
            return True
        except (ValueError, OSError):
            return False


class Response(object):
    def __init__(
        self,
        result: Optional[JSON] = None,
        id: Optional[str] = None,
        error: Optional[JSON] = None,
    ) -> None:
        self.result = result
        self.id = id
        self.error = error

    @staticmethod
    def validate_payload(payload: JSON) -> bool:
        return (
            payload.get("jsonrpc") == "2.0"
            and "result" in payload
            and "error" in payload
        )


def parse_content_length(line: bytes) -> Optional[int]:
    if line.startswith(b"Content-Length:"):
        length = line.split(b"Content-Length:")[1].strip()
        try:
            return int(length)
        except ValueError:
            return None
    return None


def _read_payload(file: BinaryIO) -> Optional[JSON]:
    try:
        line = file.readline()
        length = parse_content_length(line)
        if not length:
            return None

        # Read header lines until the empty line
        while line.strip():
            line = file.readline()

        body = file.read(length)
        return json.loads(body.decode("utf-8"))
    except (ValueError, OSError, JSONDecodeError):
        return None


def read_request(file: BinaryIO) -> Optional[Request]:
    payload = _read_payload(file)
    if payload and Request.validate_payload(payload):
        return Request(
            # pyre-fixme[6]: Expected `str` for 1st param but got `Optional[Any]`.
            method=payload.get("method"),
            id=payload.get("id"),
            parameters=payload.get("params"),
        )
    return None


def read_response(file: BinaryIO) -> Response:
    payload = _read_payload(file)
    if payload and Response.validate_payload(payload):
        return Response(
            result=payload.get("result"),
            error=payload.get("error"),
            id=payload.get("id"),
        )
    raise JSONRPCException


def perform_handshake(
    input_file: BinaryIO, output_file: BinaryIO, client_version: str
) -> None:
    server_handshake = read_request(input_file)
    if server_handshake and server_handshake.method == "handshake/server":
        server_handshake_parameters = server_handshake.parameters
        if server_handshake_parameters:
            server_version = server_handshake_parameters.get("version")
            if server_version != client_version:
                raise ValueError(
                    "Version mismatch. Server has version `{}`, "
                    "while client has version `{}`.".format(
                        server_version, client_version
                    )
                )
            client_handshake = Request(method="handshake/client")
            client_handshake.write(output_file)
        else:
            raise ValueError("Handshake parameters from server not found.")
    else:
        raise ValueError("Handshake from server was malformed.")



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import logging
import os
import re
from logging import Logger
from typing import Dict, List, Optional

from .. import json_rpc, log
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from .command import Command


LOG: Logger = logging.getLogger(__name__)


class Query(Command):
    NAME = "query"

    def _rewrite_paths(self, query: str) -> str:
        paths = re.findall(r"'[a-zA-Z_\-\.\/0-9]+\.py'", query)
        symbolic_link_mapping: Optional[Dict[str, str]] = None
        for path in paths:
            # Lazily compute the symbolic link mapping, as it can add over a second of
            # latency for large projects.
            if symbolic_link_mapping is None:
                symbolic_link_mapping = (
                    self._analysis_directory.compute_symbolic_links()
                )
            path = path[1:-1]
            absolute_path = os.path.abspath(path)
            if absolute_path in symbolic_link_mapping:
                query = query.replace(path, symbolic_link_mapping[absolute_path])
        return query

    def __init__(
        self,
        arguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Query, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self.query = self._rewrite_paths(arguments.query)

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        query_message = """
        `https://pyre-check.org/docs/querying-pyre.html` contains examples and documentation
        for this command, which queries a running pyre server for type, function and
        attribute information.

        To get a full list of queries, you can run `pyre query help`.
        """
        query = parser.add_parser(cls.NAME, epilog=query_message)
        query.set_defaults(command=cls)
        query_argument_message = """
        `pyre query help` will give a full list of available queries for \
        the running Pyre server.
         Example: `pyre query "superclasses(int)"`.
        """
        query.add_argument("query", help=query_argument_message)

    def _flags(self) -> List[str]:
        flags = [self.query]
        log_directory = self._log_directory
        if log_directory:
            flags.extend(["-log-directory", log_directory])
        return flags

    def _run(self) -> None:
        if self._configuration._use_json_sockets:
            request = json_rpc.Request(
                method="typeQuery", parameters={"query": self.query}
            )
            self._send_and_handle_socket_request(request, self._version_hash)
        else:
            result = self._call_client(command=self.NAME)
            result.check()
            log.stdout.write(result.output)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import hashlib
import json
import logging
import os
import shutil
import site
import sys
from logging import Logger
from typing import Dict, List, Optional, Union

from . import (
    BINARY_NAME,
    CONFIGURATION_FILE,
    LOG_DIRECTORY,
    assert_readable_directory,
    find_typeshed,
    number_of_workers,
)
from .exceptions import EnvironmentException


LOG: Logger = logging.getLogger(__name__)


class InvalidConfiguration(Exception):
    pass


class SearchPathElement:
    def __init__(self, root: str, subdirectory: Optional[str] = None) -> None:
        self.root = os.path.expanduser(root)
        self.subdirectory = subdirectory

    def path(self) -> str:
        subdirectory = self.subdirectory
        if subdirectory is not None:
            return os.path.join(self.root, subdirectory)
        else:
            return self.root

    def command_line_argument(self) -> str:
        subdirectory = self.subdirectory
        if subdirectory is not None:
            return self.root + "$" + subdirectory
        else:
            return self.root

    def __eq__(self, other: str) -> bool:
        # We support this for testing.
        if isinstance(other, str):
            return self.path() == other
        else:
            return self.root == other.root and self.subdirectory == other.subdirectory


def expand_search_path(path: Union[Dict[str, str], str]) -> SearchPathElement:
    if isinstance(path, str):
        return SearchPathElement(path)
    else:
        if "root" in path and "subdirectory" in path:
            root = path["root"]
            subdirectory = path["subdirectory"]
            return SearchPathElement(root, subdirectory)
        elif "site-package" in path:
            site_root = site.getsitepackages()
            subdirectory = path["site-package"]

            found_element = None
            for root in site_root:
                site_package_element = SearchPathElement(root, subdirectory)
                if os.path.isdir(site_package_element.path()):
                    found_element = site_package_element
            if found_element is None:
                raise InvalidConfiguration(
                    "Cannot find site package '{}'".format(subdirectory)
                )
            return found_element
        else:
            raise InvalidConfiguration(
                "Search path elements must have `root` and `subdirectory` specified."
            )


class _ConfigurationFile:
    def __init__(self, file) -> None:
        self._deprecated = {"do_not_check": "ignore_all_errors"}
        contents = file.read()
        self.file_hash = hashlib.sha1(contents.encode("utf-8")).hexdigest()
        self._configuration = json.loads(contents)

    def consume(
        self,
        key,
        default=None,
        current=None,
        print_on_success=False,
        raise_on_override=False,
    ):
        """
        Consume a key from the configuration. When a key is consumed, it
        is removed from the configuration.

        If not found, the default is returned. If the current value is not
        None, it will be returned instead, but the key will still be
        considered consumed.
        """

        value = self._configuration.pop(key, default)
        if raise_on_override and current and value:
            raise EnvironmentException(
                "Configuration file may not override `{}` field.".format(key)
            )
        if current:
            return current
        if value and print_on_success:
            LOG.debug("Found %s: `%s`", key, ", ".join(value))
        if value and key in self._deprecated:
            LOG.warning(
                "Configuration file uses deprecated item `%s`: "
                "please migrate to its replacement `%s`",
                key,
                self._deprecated[key],
            )
        return value

    def unused_keys(self):
        """
        Return all keys not consumed yet. Some keys are explicitly whitelisted.
        """

        return self._configuration.keys() - {
            "buck_builder_binary",
            "continuous",
            "coverage",
            "differential",
            "push_blocking",
            "saved_state",
            "taint_models_path",
        }


def expand_relative_path(root: str, path: str) -> str:
    path = os.path.expanduser(path)
    if os.path.isabs(path):
        return path
    else:
        return os.path.join(root, path)


class Configuration:
    disabled = False  # type: bool

    def __init__(
        self,
        local_configuration: Optional[str] = None,
        search_path: Optional[List[str]] = None,
        binary: Optional[str] = None,
        typeshed: Optional[str] = None,
        preserve_pythonpath: bool = False,
        excludes: Optional[List[str]] = None,
        formatter: Optional[str] = None,
        logger: Optional[str] = None,
        log_directory: Optional[str] = None,
    ) -> None:
        self.source_directories = []
        self.targets = []
        self.logger = logger
        self.formatter = formatter
        self.ignore_all_errors = []
        self.number_of_workers = None  # type: int # pyre-fixme
        self.local_configuration = None  # type: Optional[str]
        self.taint_models_path = []  # type: List[str]
        self.file_hash = None  # type: Optional[str]
        self.extensions = []  # type: List[str]
        self._log_directory = log_directory  # type: Optional[str]

        self._version_hash = None  # type: Optional[str]
        self._binary = None  # type: Optional[str]
        self._typeshed = None  # type: Optional[str]
        self.strict = False  # type: bool
        self._use_buck_builder = None  # type: Optional[bool]
        self._use_json_sockets = None  # type: Optional[bool]
        self.ignore_infer = []

        # Handle search path from multiple sources
        self._search_path = []
        if preserve_pythonpath:
            for path in os.getenv("PYTHONPATH", default="").split(":"):
                if path != "":
                    if os.path.isdir(path):
                        self._search_path.append(SearchPathElement(path))
                    else:
                        LOG.warning(
                            "`{}` is not a valid directory, dropping it "
                            "from PYTHONPATH".format(path)
                        )
            # sys.path often includes '' and a zipped python version, so
            # we don't log warnings for non-dir entries
            sys_path = [
                SearchPathElement(path) for path in sys.path if os.path.isdir(path)
            ]
            self._search_path.extend(sys_path)
        if search_path:
            search_path_elements = [expand_search_path(path) for path in search_path]
            self._search_path.extend(search_path_elements)
        # We will extend the search path further, with the config file
        # items, inside _read().

        if binary:
            self._binary = binary

        if typeshed:
            self._typeshed = typeshed

        self.excludes = []  # type: List[str]
        if excludes:
            self.excludes.extend(excludes)

        if local_configuration:
            # Handle local configuration explicitly configured on the
            # commandline.
            self._check_read_local_configuration(
                local_configuration, fail_on_error=True
            )
        if log_directory:
            self.ignore_all_errors.append(log_directory)

        self.autocomplete = False

        # Order matters. The values will only be updated if a field is None.
        self._read(CONFIGURATION_FILE)
        self._override_version_hash()
        self._resolve_versioned_paths()
        self._apply_defaults()
        self._validate()

    def _validate(self) -> None:
        try:

            def is_list_of_strings(list):
                if len(list) == 0:
                    return True
                return not isinstance(list, str) and all(
                    isinstance(element, str) for element in list
                )

            if not is_list_of_strings(
                self.source_directories
            ) or not is_list_of_strings(self.targets):
                raise InvalidConfiguration(
                    "`target` and `source_directories` fields must be lists of "
                    "strings."
                )

            if not is_list_of_strings(self.ignore_all_errors):
                raise InvalidConfiguration(
                    "`ignore_all_errors` field must be a list of strings."
                )

            if not is_list_of_strings(self.ignore_infer):
                raise InvalidConfiguration(
                    "`ignore_infer` field must be a list of strings."
                )

            if not is_list_of_strings(self.extensions):
                raise InvalidConfiguration(
                    "`extensions` field must be a list of strings."
                )
            if not all(
                extension.startswith(".") or not extension
                for extension in self.extensions
            ):
                raise InvalidConfiguration(
                    "`extensions` must only contain strings formatted as `.EXT`"
                )

            if not os.path.exists(self.binary):
                raise InvalidConfiguration(
                    "Binary at `{}` does not exist.".format(self.binary)
                )

            number_of_workers = self.number_of_workers
            if number_of_workers and number_of_workers < 1:
                raise InvalidConfiguration("Number of workers must be greater than 0.")

            # Validate typeshed path and sub-elements.
            assert_readable_directory(self.typeshed)

            # A courtesy warning since we have changed default behaviour.
            if self._typeshed_has_obsolete_value():
                LOG.warning(
                    "It appears that `{}` points at a `stdlib` directory. "
                    "Please note that the `typeshed` configuration must point at "
                    "the root of the `typeshed` directory.".format(self.typeshed)
                )

            for element in self.ignore_all_errors:
                if not os.path.exists(element):
                    LOG.warning(
                        "Nonexistent path passed in to `ignore_all_errors` \
                        field: `{}`".format(
                            element
                        )
                    )

            for element in self.ignore_infer:
                if not os.path.exists(element):
                    LOG.warning(
                        "Nonexistent path passed in to `ignore_infer` \
                        field: `{}`".format(
                            element
                        )
                    )

            typeshed_subdirectories = os.listdir(self.typeshed)
            if "stdlib" not in typeshed_subdirectories:
                raise InvalidConfiguration(
                    "`typeshed` location must contain a `stdlib` directory."
                )

            for typeshed_subdirectory_name in typeshed_subdirectories:
                typeshed_subdirectory = os.path.join(
                    self.typeshed, typeshed_subdirectory_name
                )
                if (
                    not os.path.isdir(typeshed_subdirectory)
                    or typeshed_subdirectory_name == "tests"
                    or typeshed_subdirectory_name[0] == "."
                ):
                    # Ignore some well-known directories we do not care about.
                    continue

                assert_readable_directory(typeshed_subdirectory)
                for typeshed_version_directory_name in os.listdir(
                    typeshed_subdirectory
                ):
                    if not typeshed_version_directory_name[0].isdigit():
                        raise InvalidConfiguration(
                            "Directories inside `typeshed` must only contain "
                            "second-level subdirectories starting with "
                            "a version number."
                        )
                    typeshed_version_directory = os.path.join(
                        typeshed_subdirectory, typeshed_version_directory_name
                    )
                    assert_readable_directory(typeshed_version_directory)

            # Validate elements of the search path.
            for element in self._search_path:
                assert_readable_directory(element.path())
        except InvalidConfiguration as error:
            raise EnvironmentException("Invalid configuration: {}".format(str(error)))

    @property
    def version_hash(self) -> str:
        return self._version_hash or "unversioned"

    @property
    def binary(self) -> str:
        binary = self._binary
        if not binary:
            raise InvalidConfiguration("Configuration was not validated")
        return binary

    @property
    def typeshed(self) -> str:
        typeshed = self._typeshed
        if not typeshed:
            raise InvalidConfiguration("Configuration invalid: no typeshed specified")
        return typeshed

    @property
    def use_buck_builder(self) -> bool:
        return self._use_buck_builder or False

    @property
    def search_path(self) -> List[str]:
        if not self._search_path:
            return []
        return [element.command_line_argument() for element in self._search_path]

    @property
    def local_configuration_root(self) -> Optional[str]:
        local_configuration = self.local_configuration
        if local_configuration:
            if os.path.isdir(local_configuration):
                return local_configuration
            else:
                return os.path.dirname(local_configuration)

    @property
    def log_directory(self) -> str:
        log_directory = self._log_directory
        if not log_directory:
            raise InvalidConfiguration("Configuration was not validated")
        return log_directory

    def _check_read_local_configuration(self, path: str, fail_on_error: bool) -> None:
        if fail_on_error and not os.path.exists(path):
            raise EnvironmentException(
                "Local configuration path `{}` does not exist.".format(path)
            )

        if os.path.isdir(path):
            local_configuration = os.path.join(path, CONFIGURATION_FILE + ".local")

            if not os.path.exists(local_configuration):
                if fail_on_error:
                    raise EnvironmentException(
                        "Local configuration directory `{}` does not contain "
                        "a `{}` file.".format(path, CONFIGURATION_FILE + ".local")
                    )
                else:
                    LOG.debug(
                        "Configuration will be read from the project root: "
                        "`{}`".format(os.getcwd())
                    )
            else:
                self.local_configuration = local_configuration
        else:
            local_configuration = path
            self.local_configuration = local_configuration
        self._read(local_configuration)

    def _read(self, path: str) -> None:
        try:
            with open(path) as file:
                LOG.debug("Reading configuration `%s`...", path)
                configuration = _ConfigurationFile(file)

                source_directories = configuration.consume(
                    "source_directories",
                    default=[],
                    current=self.source_directories,
                    print_on_success=True,
                    raise_on_override=True,
                )
                configuration_directory = os.path.dirname(path)
                if configuration_directory:
                    self.source_directories = [
                        os.path.join(configuration_directory, directory)
                        for directory in source_directories
                    ]
                else:
                    self.source_directories = [
                        os.path.expanduser(directory)
                        for directory in source_directories
                    ]

                self.targets = configuration.consume(
                    "targets",
                    default=[],
                    current=self.targets,
                    print_on_success=True,
                    raise_on_override=True,
                )

                if configuration.consume("disabled", default=False):
                    self.disabled = True

                self.logger = configuration.consume("logger", current=self.logger)

                self.formatter = configuration.consume(
                    "formatter", current=self.formatter
                )

                self.strict = configuration.consume("strict", default=self.strict)
                ignore_all_errors = configuration.consume(
                    "ignore_all_errors", default=[]
                )
                # Deprecated.
                ignore_all_errors += configuration.consume("do_not_check", default=[])
                configuration_path = os.path.dirname(os.path.realpath(path))
                self.ignore_all_errors.extend(
                    [
                        expand_relative_path(root=configuration_path, path=path)
                        for path in ignore_all_errors
                    ]
                )

                ignore_infer = configuration.consume("ignore_infer", default=[])
                self.ignore_infer.extend(
                    [
                        expand_relative_path(root=configuration_path, path=path)
                        for path in ignore_infer
                    ]
                )

                self.number_of_workers = int(
                    configuration.consume(
                        "workers", default=0, current=self.number_of_workers
                    )
                )

                binary = configuration.consume("binary", current=self._binary)
                assert binary is None or isinstance(binary, str)
                if binary is not None:
                    binary = expand_relative_path(configuration_path, binary)
                self._binary = binary

                additional_search_path = configuration.consume(
                    "search_path", default=[]
                )

                if isinstance(additional_search_path, list):
                    self._search_path.extend(
                        [expand_search_path(path) for path in additional_search_path]
                    )
                else:
                    self._search_path.append(SearchPathElement(additional_search_path))

                version_hash = configuration.consume(
                    "version", current=self._version_hash
                )
                assert version_hash is None or isinstance(version_hash, str)
                self._version_hash = version_hash

                typeshed = configuration.consume("typeshed", current=self._typeshed)
                assert typeshed is None or isinstance(typeshed, str)
                if typeshed is not None:
                    typeshed = expand_relative_path(configuration_path, typeshed)
                self._typeshed = typeshed

                taint_models_path = configuration.consume("taint_models_path")
                assert (
                    taint_models_path is None
                    or isinstance(taint_models_path, str)
                    or isinstance(taint_models_path, list)
                )
                configuration_directory = os.path.dirname(os.path.realpath(path))
                if isinstance(taint_models_path, str):
                    self.taint_models_path.append(
                        os.path.join(configuration_directory, taint_models_path)
                    )
                elif isinstance(taint_models_path, list):
                    self.taint_models_path.extend(
                        [
                            os.path.join(configuration_directory, path)
                            for path in taint_models_path
                        ]
                    )

                excludes = configuration.consume("exclude", default=[])
                if isinstance(excludes, list):
                    self.excludes.extend(excludes)
                else:
                    self.excludes.append(excludes)

                extensions = configuration.consume("extensions", default=[])
                self.extensions.extend(extensions)

                # We rely on the configuration SHA1 to make
                if configuration.consume("saved_state"):
                    self.file_hash = configuration.file_hash

                use_buck_builder = configuration.consume("use_buck_builder")
                if self._use_buck_builder is None:
                    self._use_buck_builder = use_buck_builder

                use_json_sockets = configuration.consume("use_json_sockets")
                if self._use_json_sockets is None:
                    self._use_json_sockets = use_json_sockets

                self.autocomplete = configuration.consume("autocomplete", default=False)

                # This block should be at the bottom to be effective.
                unused_keys = configuration.unused_keys()
                if unused_keys:
                    LOG.warning(
                        "Some configuration items were not recognized in "
                        "`{}`: {}".format(path, ", ".join(unused_keys))
                    )
        except IOError:
            # To avoid TOCTTOU bugs, handle IOErrors here silently.
            # We error elsewhere if there weren't enough parameters passed into pyre.
            pass
        except json.JSONDecodeError as error:
            raise EnvironmentException(
                "Configuration file at `{}` is invalid: {}.".format(path, str(error))
            )

    def _resolve_versioned_paths(self) -> None:
        version_hash = self.version_hash
        if not version_hash:
            return

        binary = self._binary
        if binary:
            self._binary = binary.replace("%V", version_hash)
        typeshed = self._typeshed
        if typeshed:
            self._typeshed = typeshed.replace("%V", version_hash)

    def _override_version_hash(self) -> None:
        overriding_version_hash = os.getenv("PYRE_VERSION_HASH")
        if overriding_version_hash:
            self._version_hash = overriding_version_hash
            LOG.warning("Version hash overridden with `%s`", self._version_hash)

    def _apply_defaults(self) -> None:
        overriding_binary = os.getenv("PYRE_BINARY")
        if overriding_binary:
            self._binary = overriding_binary
            LOG.warning("Binary overridden with `%s`", self._binary)
        if not self._binary:
            LOG.info(
                "No binary specified, looking for `{}` in PATH".format(BINARY_NAME)
            )
            self._binary = shutil.which(BINARY_NAME)
            if not self._binary:
                binary_candidate = os.path.join(
                    os.path.dirname(sys.argv[0]), BINARY_NAME
                )
                self._binary = shutil.which(binary_candidate)
            if not self._binary:
                LOG.warning("Could not find `{}` in PATH".format(BINARY_NAME))
            else:
                LOG.info("Found: `%s`", self._binary)

        if not self.number_of_workers:
            self.number_of_workers = number_of_workers()

        if not self._typeshed:
            LOG.info("No typeshed specified, looking for it")
            self._typeshed = find_typeshed()
            if not self._typeshed:
                LOG.warning("Could not find a suitable typeshed")
            else:
                LOG.info("Found: `%s`", self._typeshed)

        if not self._log_directory:
            # TODO(T56191177): We should not start up a server at all if no configurations
            # exist. Currently, we treat the cwd as the project root if no configurations
            # exist. Instead, we should default logging to `tmp/.pyre` in the initial
            # find_log_directory as well.
            self._log_directory = "/tmp/.pyre"

    def _typeshed_has_obsolete_value(self) -> bool:
        (head, tail) = os.path.split(self.typeshed)
        if tail == "stdlib":
            return True
        if tail != "":
            return False
        # If `path` ends in a slash, tail will be empty.
        (head, tail) = os.path.split(head)
        return tail == "stdlib"



# Copyright (c) 2019-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import functools
import logging
import multiprocessing
import os
import signal
import sys
from multiprocessing import Event
from typing import Any, Dict, List, NamedTuple

from .filesystem import acquire_lock, remove_if_exists


LOG = logging.getLogger(__name__)  # type: logging.Logger


Subscription = NamedTuple(
    "Subscription", [("root", str), ("name", str), ("subscription", Dict[str, Any])]
)


class WatchmanSubscriber(object):
    def __init__(self, base_path: str) -> None:
        self._base_path: str = base_path
        self._alive: bool = True
        self._ready: multiprocessing.synchronize.Event = Event()

    @property
    def _name(self) -> str:
        """
            A name to identify the subscriber. Used as the directory and file names
            for the log, lock, and pid files.
        """
        raise NotImplementedError

    @property
    def _subscriptions(self) -> List[Subscription]:
        """
            List of subscriptions
        """
        raise NotImplementedError

    def _handle_response(self, response: Dict[str, Any]) -> None:
        """
            Callback invoked when a message is received from watchman
        """
        raise NotImplementedError

    @property
    @functools.lru_cache(1)
    def _watchman_client(self) -> "pywatchman.client":  # noqa
        try:
            import pywatchman  # noqa

            # The client will block indefinitely when timeout is None.
            return pywatchman.client(timeout=None)
        except ImportError as exception:
            LOG.info("Not starting %s due to %s", self._name, str(exception))
            sys.exit(1)

    def _subscribe_to_watchman(self, subscription: Subscription) -> None:
        self._watchman_client.query("watch", subscription.root)
        self._watchman_client.query(
            "subscribe", subscription.root, subscription.name, subscription.subscription
        )

    def _run(self) -> None:
        try:
            os.makedirs(self._base_path)
        except OSError:
            pass
        lock_path = os.path.join(self._base_path, "{}.lock".format(self._name))
        pid_path = os.path.join(self._base_path, "{}.pid".format(self._name))

        def cleanup() -> None:
            LOG.info("Cleaning up lock and pid files before exiting.")
            remove_if_exists(pid_path)
            remove_if_exists(lock_path)

        def interrupt_handler(_signal_number=None, _frame=None) -> None:
            LOG.info("Interrupt signal received.")
            cleanup()
            sys.exit(0)

        signal.signal(signal.SIGINT, interrupt_handler)

        # Die silently if unable to acquire the lock.
        with acquire_lock(lock_path, blocking=False):
            file_handler = logging.FileHandler(
                os.path.join(self._base_path, "%s.log" % self._name), mode="w"
            )
            file_handler.setFormatter(
                logging.Formatter("%(asctime)s %(levelname)s %(message)s")
            )
            LOG.addHandler(file_handler)

            with open(pid_path, "w+") as pid_file:
                pid_file.write(str(os.getpid()))

            for subscription in self._subscriptions:
                self._subscribe_to_watchman(subscription)

            if not self._subscriptions:
                LOG.info("No watchman roots to subscribe to.")

            connection = self._watchman_client.recvConn
            if not connection:
                LOG.error("Connection to Watchman for %s not found", self._name)
                sys.exit(1)

            while self._alive:
                # This call is blocking, which prevents this loop from burning CPU.
                response = connection.receive()
                try:
                    if response["is_fresh_instance"]:
                        LOG.info(
                            "Ignoring initial watchman message for %s", response["root"]
                        )
                    else:
                        self._handle_response(response)
                    self._ready.set()  # At least one message has been received.
                except KeyError:
                    pass

            cleanup()

    def daemonize(self) -> None:
        """We double-fork here to detach the daemon process from the parent.
           If we were to just fork the child as a daemon, we'd have to worry about the
           parent process exiting zombifying the daemon."""
        if os.fork() == 0:
            pid = os.fork()
            if pid == 0:
                try:
                    LOG.propagate = False
                    # Closing the sys.stdout and stderr file descriptors here causes
                    # the program to crash when attempting to log.
                    os.close(sys.stdout.fileno())
                    os.close(sys.stderr.fileno())
                    self._run()
                    sys.exit(0)
                except Exception as exception:
                    LOG.info("Not running %s due to %s", self._name, str(exception))
                    sys.exit(1)
            else:
                sys.exit(0)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Optional

from . import upgrade_core


class VersionControl:
    @staticmethod
    def commit_message(directory: str, summary_override: Optional[str] = None) -> str:
        return ""

    @staticmethod
    def submit_changes(submit: bool, message: str) -> None:
        pass


def main() -> None:
    version_control = VersionControl()
    upgrade_core.run(version_control)


if __name__ == "__main__":
    main()



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import logging
import pathlib
import re
from logging import Logger

from .errors import errors_from_stdin, sort_errors


LOG: Logger = logging.getLogger(__name__)


def run_missing_overridden_return_annotations(
    arguments: argparse.Namespace, _version_control
) -> None:
    errors = sort_errors(errors_from_stdin(arguments))
    for path, errors in errors:
        LOG.info("Patching errors in `%s`.", path)
        errors = reversed(sorted(errors, key=lambda error: error["line"]))

        path = pathlib.Path(path)
        lines = path.read_text().split("\n")

        for error in errors:
            if error["code"] != 15:
                continue
            line = error["line"] - 1

            match = re.match(r".*`(.*)`\.", error["description"])
            if not match:
                continue
            annotation = match.groups()[0]

            # Find last closing parenthesis in after line.
            LOG.info("Looking at %d: %s", line, lines[line])
            while True:
                if "):" in lines[line]:
                    lines[line] = lines[line].replace("):", ") -> %s:" % annotation)
                    LOG.info("%d: %s", line, lines[line])
                    break
                else:
                    line = line + 1

        LOG.warn("Writing patched %s", str(path))
        path.write_text("\n".join(lines))


def run_missing_global_annotations(
    arguments: argparse.Namespace, _version_control
) -> None:
    errors = sort_errors(errors_from_stdin(arguments))
    for path, errors in errors:
        LOG.info("Patching errors in `%s`", path)
        errors = reversed(sorted(errors, key=lambda error: error["line"]))

        path = pathlib.Path(path)
        lines = path.read_text().split("\n")

        for error in errors:
            if error["code"] != 5:
                continue
            line = error["line"] - 1

            match = re.match(r".*`.*`.*`(.*)`.*", error["description"])
            if not match:
                continue
            annotation = match.groups()[0]

            LOG.info("Looking at %d: %s", line, lines[line])
            if " =" in lines[line]:
                lines[line] = lines[line].replace(" =", ": %s =" % annotation)
                LOG.info("%d: %s", line, lines[line])

        path.write_text("\n".join(lines))



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import logging
import subprocess
from logging import Logger


LOG: Logger = logging.getLogger(__name__)


def get_lint_status() -> int:
    lint_status = subprocess.run(
        [
            "arc",
            "lint",
            "--never-apply-patches",
            "--enforce-lint-clean",
            "--output",
            "none",
        ]
    )
    return lint_status.returncode


def apply_lint() -> None:
    LOG.info("Lint was dirty after file modifications. Cleaning lint and re-checking.")
    subprocess.run(["arc", "lint", "--apply-patches", "--output", "none"])



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import json
import logging
import multiprocessing
import os
import platform
import subprocess
import sys
import time
import traceback
from argparse import Namespace
from pathlib import Path
from typing import Any, Dict, Optional, Set, TextIO

from . import buck
from .exceptions import EnvironmentException
from .filesystem import find_root, translate_paths  # noqa


CONFIGURATION_FILE: str = ".pyre_configuration"
BINARY_NAME: str = "pyre.bin"
CLIENT_NAME: str = "pyre-client"
LOG_DIRECTORY: str = ".pyre"


LOG = logging.getLogger(__name__)  # type: logging.Logger


def assert_readable_directory(directory: str) -> None:
    if not os.path.isdir(directory):
        raise EnvironmentException("{} is not a valid directory.".format(directory))
    if not os.access(directory, os.R_OK):
        raise EnvironmentException("{} is not a readable directory.".format(directory))


def assert_writable_directory(directory: str) -> None:
    if not os.path.isdir(directory):
        raise EnvironmentException("{} is not a valid directory.".format(directory))
    if not os.access(directory, os.W_OK):
        raise EnvironmentException("{} is not a writable directory.".format(directory))


def readable_directory(directory: str) -> str:
    assert_readable_directory(directory)
    return directory


def is_capable_terminal(file: TextIO = sys.stderr) -> bool:
    """
    Determine whether we are connected to a capable terminal.
    """
    if not os.isatty(file.fileno()):
        return False
    terminal = os.getenv("TERM", "dumb")
    # Hardcoded list of non-capable terminals.
    return terminal not in ["dumb", "emacs"]


def get_binary_version(configuration) -> str:
    override = os.getenv("PYRE_BINARY")
    if override:
        return "override: {}".format(override)

    configured = configuration.version_hash
    if configured:
        return configured

    return "No version set"


def get_binary_version_from_file(local_path: Optional[str]) -> str:
    override = os.getenv("PYRE_BINARY")
    if override:
        return "override: {}".format(override)

    def read_version(configuration_path: str) -> Optional[str]:
        with open(configuration_path) as file:
            configuration_contents = file.read()
            return json.loads(configuration_contents).pop("version", None)

    version = None  # type: Optional[str]
    try:
        # Get local configuration version
        if local_path:
            local_configuration = os.path.join(
                local_path, CONFIGURATION_FILE + ".local"
            )
            version = read_version(local_configuration)

        # Get configuration version
        if not version:
            version = read_version(CONFIGURATION_FILE)
    except Exception:
        pass
    # pyre-fixme[7]: Expected `str` but got `Optional[str]`.
    return "No version set" if not version else version


def find_project_root(original_directory: str) -> str:
    """Pyre always runs from the directory containing the nearest .pyre_configuration,
    if one exists."""
    global_root = find_root(original_directory, CONFIGURATION_FILE)
    return global_root or original_directory


def find_local_root(original_directory: str) -> Optional[str]:
    global_root = find_root(original_directory, CONFIGURATION_FILE)
    local_root = find_root(original_directory, CONFIGURATION_FILE + ".local")
    # Check for illegal nested local configuration.
    if local_root:
        parent_local_root = find_root(
            os.path.dirname(local_root), CONFIGURATION_FILE + ".local"
        )
        if parent_local_root:
            raise EnvironmentException(
                "Local configuration is nested under another local configuration at "
                "`{}`. Please combine the sources into a single configuration or split "
                "the parent configuration to avoid inconsistent errors.".format(
                    parent_local_root
                )
            )

    # If the global configuration root is deeper than local configuration, ignore local.
    if global_root and local_root and global_root.startswith(local_root):
        local_root = None
    if local_root:
        return local_root


def find_log_directory(
    log_directory: Optional[str],
    current_directory: str,
    local_configuration: Optional[str],
) -> str:
    """Pyre outputs all logs to a .pyre directory that lives in the project root."""
    if not log_directory:
        log_directory = os.path.join(current_directory, LOG_DIRECTORY)
        if local_configuration:
            # `log_directory` will never escape `.pyre/` because in `switch_root` we have
            # guaranteed that configurations are never deeper than local configurations
            relative = os.path.relpath(local_configuration, current_directory)
            log_directory = os.path.join(log_directory, relative)
    Path(log_directory).mkdir(parents=True, exist_ok=True)
    return log_directory


def _resolve_filter_paths(
    arguments: Namespace, configuration, original_directory: str
) -> Set[str]:
    filter_paths = []
    if arguments.source_directories or arguments.targets:
        if arguments.source_directories:
            filter_paths += arguments.source_directories
        if arguments.targets:
            filter_paths += [
                buck.presumed_target_root(target) for target in arguments.targets
            ]
    else:
        local_configuration_root = configuration.local_configuration_root
        if local_configuration_root:
            filter_paths = [local_configuration_root]
    return translate_paths(filter_paths, original_directory)


def number_of_workers() -> int:
    try:
        return max(multiprocessing.cpu_count() - 4, 1)
    except NotImplementedError:
        return 4


def log_statistics(
    category: str,
    arguments: Optional[Namespace] = None,
    # this is typed as a Any because configuration imports __init__
    configuration: Optional[Any] = None,
    integers: Optional[Dict[str, int]] = None,
    normals: Optional[Dict[str, str]] = None,
    logger: Optional[str] = None,
) -> None:
    integers = integers or {}
    if "time" not in integers:
        integers["time"] = int(time.time())
    normals = normals or {}
    if configuration:
        normals = {
            **normals,
            "version": configuration.version_hash,
        }  # type: Dict[str, str]
        if not logger:
            logger = configuration.logger
    if not logger:
        raise ValueError("Logger must either be given or in configuration")
    if arguments:
        normals = {**normals, "arguments": str(arguments)}
    try:
        statistics = {
            "int": integers,
            "normal": {
                **normals,
                "command_line": " ".join(sys.argv),
                "host": platform.node() or "",
                "platform": platform.system() or "",
                "user": os.getenv("USER", ""),
            },
        }
        statistics = json.dumps(statistics).encode("ascii", "strict")
        subprocess.run([logger, category], input=statistics)
    except Exception:
        LOG.warning("Unable to log using `%s`", logger)
        LOG.info(traceback.format_exc())


def _find_directory_upwards(base: str, target: str) -> Optional[str]:
    """
    Walk directories upwards from base, until the root directory is
    reached. At each step, check if the target directory exist, and return
    it if found. Return None if the search is unsuccessful.
    """
    while True:
        step = os.path.join(base, target)
        LOG.debug("Trying with: `%s`", step)
        if os.path.isdir(step):
            return step
        parent_directory = os.path.dirname(base)
        if parent_directory == base:
            # We have reached the root.
            break
        base = parent_directory
    return None


def find_typeshed() -> Optional[str]:
    override = os.getenv("PYRE_TYPESHED")
    if override:
        return override

    current_directory = os.path.dirname(os.path.realpath(__file__))

    # Prefer the typeshed we bundled ourselves (if any) to the one
    # from the environment.
    bundled_typeshed = _find_directory_upwards(
        current_directory, "pyre_check/typeshed/"
    )
    if bundled_typeshed:
        return bundled_typeshed

    try:
        import typeshed  # pyre-fixme: Can't find module import typeshed

        return typeshed.typeshed
    except ImportError:
        LOG.debug("`import typeshed` failed, attempting a manual lookup")

    # This is a terrible, terrible hack.
    return _find_directory_upwards(current_directory, "typeshed/")


def find_taint_models_directory() -> Optional[str]:
    return _find_directory_upwards(
        os.path.dirname(os.path.realpath(__file__)), "pyre_check/taint/"
    )



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import json
import os
from collections import defaultdict
from pathlib import Path
from re import compile
from typing import Dict, List, Optional, Pattern, Sequence

import libcst as cst

from .. import log, log_statistics
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from .command import Command


class StatisticsCollector(cst.CSTVisitor):
    def build_json(self) -> Dict[str, int]:
        return {}


class AnnotationCountCollector(StatisticsCollector):
    def __init__(
        self,
        return_count: int = 0,
        annotated_return_count: int = 0,
        globals_count: int = 0,
        annotated_globals_count: int = 0,
        parameter_count: int = 0,
        annotated_parameter_count: int = 0,
        attribute_count: int = 0,
        annotated_attribute_count: int = 0,
    ) -> None:
        self.return_count = return_count
        self.annotated_return_count = annotated_return_count
        self.globals_count = globals_count
        self.annotated_globals_count = annotated_globals_count
        self.parameter_count = parameter_count
        self.annotated_parameter_count = annotated_parameter_count
        self.attribute_count = attribute_count
        self.annotated_attribute_count = annotated_attribute_count
        self.in_class_definition = False
        self.in_function_definition = False
        self.is_static_function = False

    def build_json(self) -> Dict[str, int]:
        return {
            "return_count": self.return_count,
            "annotated_return_count": self.annotated_return_count,
            "globals_count": self.globals_count,
            "annotated_globals_count": self.annotated_globals_count,
            "parameter_count": self.parameter_count,
            "annotated_parameter_count": self.annotated_parameter_count,
            "attribute_count": self.attribute_count,
            "annotated_attribute_count": self.annotated_attribute_count,
        }

    def _is_self(self, parameter: cst.Param) -> bool:
        return (
            parameter.name.value == "self"
            and self.in_class_definition
            and not self.is_static_function
        )

    def _check_parameter_annotations(self, parameters: Sequence[cst.Param]) -> None:
        for parameter in list(parameters):
            self.parameter_count += 1
            annotation = parameter.annotation
            if annotation is not None or self._is_self(parameter):
                self.annotated_parameter_count += 1

    def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
        for decorator in node.decorators:
            decorator_node = decorator.decorator
            if isinstance(decorator_node, cst.Name):
                if decorator_node.value == "staticmethod":
                    self.is_static_function = True
        self.in_function_definition = True
        self.return_count += 1
        if node.returns is not None:
            self.annotated_return_count += 1

        self._check_parameter_annotations(node.params.default_params)
        self._check_parameter_annotations(node.params.params)

    def leave_FunctionDef(self, node: cst.FunctionDef) -> None:
        self.in_function_definition = False
        self.is_static_function = False

    def visit_Assign(self, node: cst.Assign) -> None:
        if self.in_function_definition:
            return
        if self.in_class_definition:
            self.attribute_count += 1
        else:
            self.globals_count += 1

    def visit_AnnAssign(self, node: cst.AnnAssign) -> None:
        if self.in_function_definition:
            return
        if self.in_class_definition:
            self.attribute_count += 1
            self.annotated_attribute_count += 1
        else:
            self.globals_count += 1
            self.annotated_globals_count += 1

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        self.in_class_definition = True

    def leave_ClassDef(self, node: cst.ClassDef) -> None:
        self.in_class_definition = False


class CountCollector(StatisticsCollector):
    def __init__(self, regex: str) -> None:
        self.counts: Dict[str, int] = defaultdict(int)
        self.regex: Pattern[str] = compile(regex)

    def visit_Comment(self, node: cst.Comment) -> None:
        match = self.regex.match(node.value)
        if match:
            self.counts[match.group(1)] += 1

    def build_json(self) -> Dict[str, int]:
        return dict(self.counts)


class FixmeCountCollector(CountCollector):
    def __init__(self) -> None:
        super().__init__(r"# pyre-fixme\[(\d*)\]:")


class IgnoreCountCollector(CountCollector):
    def __init__(self) -> None:
        super().__init__(r"# pyre-ignore\[(\d*)\]:")


class StrictCountCollector(StatisticsCollector):
    def __init__(self, strict_by_default: bool) -> None:
        self.is_strict: bool = False
        self.is_unsafe: bool = False
        self.strict_count: int = 0
        self.unsafe_count: int = 0
        self.strict_by_default: bool = strict_by_default
        self.unsafe_regex: Pattern[str] = compile(r"# pyre-unsafe")
        self.strict_regex: Pattern[str] = compile(r"# pyre-strict")

    def visit_Module(self, node: cst.Module) -> None:
        self.is_strict = False
        self.is_unsafe = False

    def visit_Comment(self, node: cst.Comment) -> None:
        strict_match = self.strict_regex.match(node.value)
        if strict_match:
            self.is_strict = True
        unsafe_match = self.unsafe_regex.match(node.value)
        if unsafe_match:
            self.is_unsafe = True

    def leave_Module(self, node: cst.Module) -> None:
        if self.is_unsafe:
            self.unsafe_count += 1
        elif self.is_strict or self.strict_by_default:
            self.strict_count += 1
        else:
            self.unsafe_count += 1

    def build_json(self) -> Dict[str, int]:
        return {"unsafe_count": self.unsafe_count, "strict_count": self.strict_count}


def _get_paths(target_directory: Path) -> List[Path]:
    return [
        path
        for path in target_directory.glob("**/*.py")
        if not path.name.startswith("__") and not path.name.startswith(".")
    ]


def _parse_directory(directory: Path) -> List[cst.Module]:
    files = _get_paths(directory)
    new_files = []
    for file in files:
        new_files.append(_parse_file(file))
    return new_files


def _parse_file(path: Path) -> cst.Module:
    return cst.parse_module(path.read_text())


def _parse_paths(paths: List[Path]) -> List[cst.Module]:
    parsed_paths = []
    for path in paths:
        if path.is_dir():
            parsed_directory_paths = _parse_directory(path)
            for path in parsed_directory_paths:
                parsed_paths.append(path)
        else:
            parsed_paths.append(_parse_file(path))
    return parsed_paths


def _count(
    paths: List[cst.Module], collector: StatisticsCollector
) -> StatisticsCollector:
    for path in paths:
        path.visit(collector)
    return collector


def _pyre_configuration_directory(local_configuration: Optional[str]) -> Path:
    if local_configuration:
        return Path(local_configuration.replace(".pyre_configuration.local", ""))
    return Path.cwd()


def _find_paths(local_configuration: Optional[str], paths: List[str]) -> List[Path]:
    pyre_configuration_directory = _pyre_configuration_directory(local_configuration)

    if paths:
        return [pyre_configuration_directory / path for path in paths]
    return [pyre_configuration_directory]


def file_exists(path: str) -> str:
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError("ERROR: " + str(path) + " does not exist")
    return path


class Statistics(Command):
    NAME = "statistics"

    def __init__(
        self,
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Statistics, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self._filter_paths: List[str] = arguments.filter_paths
        self._strict: bool = self._configuration.strict

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        statistics = parser.add_parser(cls.NAME)
        statistics.set_defaults(command=cls)
        statistics.add_argument(
            "filter_paths",
            nargs="*",
            type=file_exists,
            help="Source path(s) to gather metrics for.",
        )

    def _run(self) -> None:
        self._analysis_directory.prepare()
        paths = _find_paths(self._local_configuration, self._filter_paths)
        parsed_paths = _parse_paths(paths)
        annotations = _count(parsed_paths, AnnotationCountCollector())
        fixmes = _count(parsed_paths, FixmeCountCollector())
        ignores = _count(parsed_paths, IgnoreCountCollector())
        strict_files = _count(parsed_paths, StrictCountCollector(self._strict))
        data = {
            "annotations": annotations.build_json(),
            "fixmes": fixmes.build_json(),
            "ignores": ignores.build_json(),
            "strict": strict_files.build_json(),
        }
        log.stdout.write(json.dumps(data))

        if self._configuration and self._configuration.logger:
            root = str(_pyre_configuration_directory(self._local_configuration))
            log_statistics(
                "perfpipe_pyre_annotation_counts",
                configuration=self._configuration,
                integers=data["annotations"],
                normals={"root": root},
            )
            self._log_fixmes("fixme", data["fixmes"], root)
            self._log_fixmes("ignore", data["ignores"], root)
            log_statistics(
                "perfpipe_pyre_strict_adoption",
                configuration=self._configuration,
                integers=data["strict"],
                normals={"root": root},
            )

    def _log_fixmes(self, fixme_type: str, data: Dict[str, int], root: str) -> None:
        for error_code, count in data.items():
            log_statistics(
                "perfpipe_pyre_fixme_counts",
                configuration=self._configuration,
                integers={"count": count},
                normals={"root": root, "code": error_code, "type": fixme_type},
            )



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import errno
import json
import logging
import os
from logging import Logger
from typing import List, Optional

from .. import configuration_monitor, filesystem, project_files_monitor
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from .command import ExitCode, IncrementalStyle, typeshed_search_path
from .reporting import Reporting


LOG: Logger = logging.getLogger(__name__)


def _fine_grained_incremental_override(feature_string: str) -> bool:
    try:
        return json.loads(feature_string).get("enable_fine_grained_incremental", False)
    except Exception:
        return False


class Start(Reporting):
    NAME = "start"

    def __init__(
        self,
        arguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Start, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self._terminal: bool = arguments.terminal
        self._store_type_check_resolution: bool = arguments.store_type_check_resolution
        self._use_watchman: bool = not arguments.no_watchman

        features = self._features
        fine_grained_override = (
            _fine_grained_incremental_override(features)
            if features is not None
            else False
        )
        default_incremental_style = (
            IncrementalStyle.FINE_GRAINED
            if fine_grained_override
            else IncrementalStyle.SHALLOW
        )
        self._incremental_style: IncrementalStyle = (
            arguments.incremental_style or default_incremental_style
        )

        if self._no_saved_state:
            self._save_initial_state_to: Optional[str] = None
            self._changed_files_path: Optional[str] = None
            self._load_initial_state_from: Optional[str] = None
            self._saved_state_project: Optional[str] = None

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        start = parser.add_parser(cls.NAME, epilog="Starts a pyre server as a daemon.")
        start.set_defaults(command=cls)
        start.add_argument(
            "--terminal", action="store_true", help="Run the server in the terminal."
        )
        start.add_argument(
            "--store-type-check-resolution",
            action="store_true",
            help="Store extra information for `types` queries.",
        )
        start.add_argument(
            "--no-watchman",
            action="store_true",
            help="Do not spawn a watchman client in the background.",
        )
        start.add_argument(
            "--incremental-style",
            type=IncrementalStyle,
            choices=list(IncrementalStyle),
            default=None,
            help="How to approach doing incremental checks.",
        )

    def _run(self) -> None:
        blocking = False
        lock = os.path.join(self._log_directory, "client.lock")
        while True:
            # Be optimistic in grabbing the lock in order to provide users with
            # a message when the lock is being waited on.
            try:
                with filesystem.acquire_lock(lock, blocking):
                    configuration_monitor.ConfigurationMonitor(
                        self._arguments,
                        self._configuration,
                        self._analysis_directory,
                        self._current_directory,
                    ).daemonize()

                    # This unsafe call is OK due to the client lock always
                    # being acquired before starting a server - no server can
                    # spawn in the interim which would cause a race.
                    try:
                        with filesystem.acquire_lock(
                            os.path.join(self._log_directory, "server", "server.lock"),
                            blocking=False,
                        ):
                            pass
                    except OSError:
                        LOG.warning(
                            "Server at `%s` exists, skipping.",
                            self._analysis_directory.get_root(),
                        )
                        self._exit_code = ExitCode.FAILURE
                        return

                    self._analysis_directory.prepare()

                    self._call_client(command=self.NAME).check()

                    if self._use_watchman:
                        try:
                            file_monitor = project_files_monitor.ProjectFilesMonitor(  # noqa
                                self._configuration,
                                self._current_directory,
                                self._analysis_directory,
                            )
                            file_monitor.daemonize()
                            LOG.info("Initialized file monitor.")
                        except project_files_monitor.MonitorException as error:
                            LOG.warning("Failed to initialize file monitor: %s", error)

                    return
            except OSError as exception:
                if exception.errno == errno.EAGAIN:
                    blocking = True
                    LOG.info("Waiting on the pyre client lock.")
                else:
                    raise exception

    def _flags(self) -> List[str]:
        flags = super()._flags()
        if self._taint_models_path:
            for path in self._taint_models_path:
                flags.extend(["-taint-models", path])
        filter_directories = self._get_directories_to_analyze()
        if len(filter_directories):
            flags.extend(["-filter-directories", ";".join(sorted(filter_directories))])
        if len(self._configuration.ignore_all_errors):
            flags.extend(
                [
                    "-ignore-all-errors",
                    ";".join(sorted(self._configuration.ignore_all_errors)),
                ]
            )
        if self._terminal:
            flags.append("-terminal")
        if self._store_type_check_resolution:
            flags.append("-store-type-check-resolution")
        save_initial_state_to = self._save_initial_state_to
        if save_initial_state_to and os.path.isdir(
            os.path.dirname(save_initial_state_to)
        ):
            flags.extend(["-save-initial-state-to", save_initial_state_to])
        saved_state_project = self._saved_state_project
        if saved_state_project:
            flags.extend(["-saved-state-project", saved_state_project])
            local_configuration_root = self._configuration.local_configuration_root
            if local_configuration_root is not None:
                relative = os.path.relpath(local_configuration_root)
                flags.extend(["-saved-state-metadata", relative.replace("/", "$")])
        configuration_file_hash = self._configuration.file_hash
        if configuration_file_hash:
            flags.extend(["-configuration-file-hash", configuration_file_hash])
        load_initial_state_from = self._load_initial_state_from
        if load_initial_state_from is not None:
            flags.extend(["-load-state-from", load_initial_state_from])
            changed_files_path = self._changed_files_path
            if changed_files_path is not None:
                flags.extend(["-changed-files-path", changed_files_path])
        elif self._changed_files_path is not None:
            LOG.error(
                "--load-initial-state-from must be set if --changed-files-path is set."
            )
        flags.extend(
            [
                "-workers",
                str(self._number_of_workers),
                "-expected-binary-version",
                self._configuration.version_hash,
            ]
        )
        search_path = self._configuration.search_path + typeshed_search_path(
            self._configuration.typeshed
        )
        if search_path:
            flags.extend(["-search-path", ",".join(search_path)])

        excludes = self._configuration.excludes
        for exclude in excludes:
            flags.extend(["-exclude", exclude])

        extensions = self._configuration.extensions
        for extension in extensions:
            flags.extend(["-extension", extension])

        if self._incremental_style == IncrementalStyle.TRANSITIVE:
            flags.append("-transitive")
        elif self._incremental_style == IncrementalStyle.FINE_GRAINED:
            flags.append("-new-incremental-check")

        if self._configuration.autocomplete:
            flags.append("-autocomplete")
        flags.extend(self._feature_flags())

        return flags



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import logging
import os
import subprocess
import sys
import time
from argparse import Namespace
from logging import Logger
from typing import Iterable

from . import CONFIGURATION_FILE, FAILURE, SUCCESS, log
from .configuration import Configuration
from .exceptions import EnvironmentException
from .filesystem import find_root


LOG: Logger = logging.getLogger(__name__)


def _parallel_check(command: Iterable[str], process_count) -> float:
    LOG.info(
        "Running %d process%s of `%s`",
        process_count,
        "es" if process_count > 1 else "",
        " ".join(command),
    )
    processes = []
    start = time.time()
    for _ in range(process_count):
        processes.append(
            subprocess.Popen(
                command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )
        )
    for process in processes:
        process.wait()
    return time.time() - start


def _compare_parallel_check(
    arguments: argparse.Namespace, configuration: Configuration, project_root: str
) -> None:
    if not os.path.isdir(arguments.source_directory):
        raise EnvironmentException(
            "`{}` is not a valid source directory.".format(arguments.source_directory)
        )
    flags = ["-typeshed", configuration.typeshed, "-project-root", project_root]
    search_path = configuration.search_path
    if search_path:
        flags.extend(["-search-path", ",".join(search_path)])
    client_command = [configuration.binary, "check"]
    client_command.extend(flags)
    client_command.append(arguments.source_directory)

    process_count = arguments.min
    while process_count < arguments.max:
        time_elapsed = _parallel_check(client_command, process_count)  # ms
        time_elapsed_per_process = time_elapsed / process_count
        LOG.info(
            "Ran %d concurrent `pyre check` process%s in %dm%ds: "
            + "%dm%ds per process.",
            process_count,
            "es" if process_count > 1 else "",
            time_elapsed / 60,
            time_elapsed % 60,
            time_elapsed_per_process / 60,
            time_elapsed_per_process % 60,
        )
        process_count += 1


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--source-directory", action="store", help="Source directory to run check on."
    )
    parser.add_argument(
        "--min",
        action="store",
        default=1,
        help="Minimum number of concurrent processes to measure.",
    )
    parser.add_argument(
        "--max",
        action="store",
        default=10,
        help="Maximum number of concurrent processes to measure.",
    )
    arguments: Namespace = parser.parse_args()
    arguments: Namespace = True
    log.initialize(noninteractive=False)

    try:
        exit_code = SUCCESS
        root = find_root(os.getcwd(), CONFIGURATION_FILE) or os.getcwd()
        os.chdir(root)
        configuration = Configuration(local_root=arguments.local_root)
        _compare_parallel_check(arguments, configuration, root)
    except Exception as error:
        LOG.error(str(error))
        exit_code = FAILURE
    finally:
        log.cleanup()
        sys.exit(exit_code)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import atexit
import json
import logging
import os
import subprocess
from logging import Logger
from typing import IO, List, Optional, cast

from .. import json_rpc
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from ..project_files_monitor import MonitorException, ProjectFilesMonitor
from .command import (
    ClientException,
    ExitCode,
    IncrementalStyle,
    Result,
    State,
    typeshed_search_path,
)
from .reporting import Reporting
from .start import Start


LOG: Logger = logging.getLogger(__name__)


def _convert_to_result(response: json_rpc.Response):
    error_code = ExitCode.FAILURE if response.error else ExitCode.SUCCESS
    return Result(output=json.dumps(response.result), code=error_code)


class Incremental(Reporting):
    NAME = "incremental"

    def __init__(
        self,
        arguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Incremental, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self._nonblocking = arguments.nonblocking  # type: bool
        self._incremental_style = arguments.incremental_style  # type: bool
        self._no_start_server = arguments.no_start  # type: bool

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        incremental_help = """
        Connects to a running Pyre server and returns the current type errors for your
        project. If no server exists for your projects, starts a new one. Running `pyre`
        implicitly runs `pyre incremental`.

        By default, incremental checks ensure that all dependencies of changed files are
        analyzed before returning results. If you'd like to get partial type checking
        results eagerly, you can run `pyre incremental --nonblocking`.
        """
        incremental = parser.add_parser(cls.NAME, epilog=incremental_help)
        incremental.set_defaults(command=cls)
        incremental.add_argument(
            "--nonblocking",
            action="store_true",
            help=(
                "Ask the server to return partial results immediately, "
                "even if analysis is still in progress."
            ),
        )
        incremental.add_argument(
            "--incremental-style",
            type=IncrementalStyle,
            choices=list(IncrementalStyle),
            default=None,
            help="How to approach doing incremental checks.",
        )
        incremental.add_argument(
            "--no-start", action="store_true", help=argparse.SUPPRESS
        )

    def _run(self) -> None:
        if (not self._no_start_server) and self._state() == State.DEAD:
            LOG.warning("Starting server at `%s`.", self._analysis_directory.get_root())
            arguments = self._arguments
            # pyre-fixme[16]: `Namespace` has no attribute `terminal`.
            arguments.terminal = False
            # pyre-fixme[16]: `Namespace` has no attribute
            #  `store_type_check_resolution`.
            arguments.store_type_check_resolution = False
            # pyre-fixme[16]: `Namespace` has no attribute `no_watchman`.
            arguments.no_watchman = False
            exit_code = (
                Start(
                    arguments,
                    self._original_directory,
                    self._configuration,
                    self._analysis_directory,
                )
                .run()
                .exit_code()
            )
            if exit_code != ExitCode.SUCCESS:
                self._exit_code = ExitCode.FAILURE
                return
        else:
            self._refresh_file_monitor()

        if self._state() != State.DEAD:
            LOG.info("Waiting for server...")

        if self._configuration._use_json_sockets:
            request = json_rpc.Request(
                method="displayTypeErrors", parameters={"files": []}
            )
            self._send_and_handle_socket_request(request, self._version_hash)
        else:
            result = self._call_client(command=self.NAME)
            try:
                result.check()
                errors = self._get_errors(result)
                self._print(errors)

                if errors:
                    self._exit_code = ExitCode.FOUND_ERRORS
            except ClientException as exception:
                LOG.error("Error while waiting for server: %s", str(exception))
                LOG.error("Run `pyre restart` in order to restart the server.")
                self._exit_code = ExitCode.FAILURE

    def _socket_result_handler(self, result: Result) -> None:
        errors = self._get_errors(result)
        self._print(errors)
        if errors:
            self._exit_code = ExitCode.FOUND_ERRORS

    def _flags(self) -> List[str]:
        flags = super()._flags()
        flags.extend(["-expected-binary-version", self._configuration.version_hash])

        search_path = self._configuration.search_path + typeshed_search_path(
            self._configuration.typeshed
        )
        if search_path:
            flags.extend(["-search-path", ",".join(search_path)])

        excludes = self._configuration.excludes
        for exclude in excludes:
            flags.extend(["-exclude", exclude])

        extensions = self._configuration.extensions
        for extension in extensions:
            flags.extend(["-extension", extension])

        if self._nonblocking:
            flags.append("-nonblocking")

        if self._incremental_style == IncrementalStyle.TRANSITIVE:
            flags.append("-transitive")

        return flags

    def _read_stderr(self, _stream) -> None:
        stderr_file = os.path.join(self._log_directory, "server/server.stdout")
        with subprocess.Popen(
            ["tail", "--follow", "--lines=0", stderr_file],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        ) as stderr_tail:
            atexit.register(stderr_tail.terminate)
            super(Incremental, self)._read_stderr(cast(IO[bytes], stderr_tail.stdout))

    def _refresh_file_monitor(self) -> None:
        if not ProjectFilesMonitor.is_alive(self._configuration):
            LOG.info("File monitor is not running.")
            try:
                ProjectFilesMonitor(
                    self._configuration,
                    self._current_directory,
                    self._analysis_directory,
                ).daemonize()
                LOG.info("Restarted file monitor.")
            except MonitorException as exception:
                LOG.warning("Failed to restart file monitor: %s", exception)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe


import argparse
import json
import logging
import re
import subprocess
import sys
import traceback
from collections import defaultdict
from logging import Logger
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from ...client.commands import ExitCode
from ...client.filesystem import get_filesystem
from .ast import verify_stable_ast
from .codemods import (
    run_missing_global_annotations,
    run_missing_overridden_return_annotations,
)
from .errors import errors_from_stdin, filter_errors, json_to_errors, sort_errors
from .postprocess import apply_lint, get_lint_status


LOG: Logger = logging.getLogger(__name__)


class Configuration:
    def __init__(self, path: Path, json_contents: Dict[str, Any]) -> None:
        self._path = path
        if path.name == ".pyre_configuration.local":
            self.is_local = True
        else:
            self.is_local = False
        self.root = str(path.parent)
        self.targets = json_contents.get("targets")
        self.source_directories = json_contents.get("source_directories")
        self.push_blocking = bool(json_contents.get("push_blocking"))
        self.version = json_contents.get("version")

    @staticmethod
    def find_project_configuration(directory: Optional[Path] = None) -> Optional[Path]:
        directory = directory or Path.cwd()
        root = directory.root
        while directory != root:
            configuration_path = directory / ".pyre_configuration"
            if configuration_path.is_file():
                return configuration_path
            directory = directory.parent
        return None

    @staticmethod
    def gather_local_configurations(arguments) -> List["Configuration"]:
        LOG.info("Finding configurations...")
        configuration_paths = [
            Path(path)
            for path in get_filesystem().list(
                ".", patterns=[r"**\.pyre_configuration.local"]
            )
        ]
        if not configuration_paths:
            LOG.info("No projects with local configurations found.")
            project_configuration = Configuration.find_project_configuration()
            if project_configuration:
                configuration_paths = [project_configuration]
            else:
                LOG.error("No project configuration found.")
                return []
        configurations = []
        for configuration_path in configuration_paths:
            with open(configuration_path) as configuration_file:
                try:
                    configuration = Configuration(
                        configuration_path, json.load(configuration_file)
                    )
                    if configuration.push_blocking or (
                        not arguments.push_blocking_only
                    ):
                        configurations.append(configuration)
                except json.decoder.JSONDecodeError:
                    LOG.error(
                        "Configuration at `%s` is invalid, skipping.",
                        configuration_path,
                    )
        LOG.info(
            "Found %d %sconfiguration%s",
            len(configurations),
            "push-blocking " if arguments.push_blocking_only else "",
            "s" if len(configurations) != 1 else "",
        )
        return configurations

    def get_path(self) -> Path:
        return self._path

    def get_directory(self) -> Path:
        return self._path.parent

    def remove_version(self) -> None:
        with open(self._path) as configuration_file:
            contents = json.load(configuration_file)
        if "version" not in contents:
            LOG.info("Version not found in configuration.")
            return
        del contents["version"]
        with open(self._path, "w") as configuration_file:
            json.dump(contents, configuration_file, sort_keys=True, indent=2)
            configuration_file.write("\n")

    def add_strict(self) -> None:
        with open(self._path) as configuration_file:
            contents = json.load(configuration_file)
        if "strict" in contents:
            LOG.info("Configuration is already strict.")
            return
        contents["strict"] = True
        with open(self._path, "w") as configuration_file:
            json.dump(contents, configuration_file, sort_keys=True, indent=2)
            configuration_file.write("\n")

    def get_errors(self, should_clean: bool = True) -> List[Dict[str, Any]]:
        # TODO(T37074129): Better parallelization or truncation needed for fbcode
        if self.targets and should_clean:
            try:
                # If building targets, run clean or space may run out on device!
                LOG.info("Running `buck clean`...")
                subprocess.call(["buck", "clean"], timeout=200)
            except subprocess.TimeoutExpired:
                LOG.warning("Buck timed out. Try running `buck kill` before retrying.")
                return []
            except subprocess.CalledProcessError as error:
                LOG.warning("Error calling `buck clean`: %s", str(error))
                return []
        try:
            LOG.info("Checking `%s`...", self.root)
            if self.is_local:
                process = subprocess.run(
                    ["pyre", "-l", self.root, "--output=json", "check"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            else:
                process = subprocess.run(
                    ["pyre", "--output=json", "check"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            json_string = process.stdout.decode().strip()
            errors = json_to_errors(json_string)
            LOG.info("Found %d error%s.", len(errors), "s" if len(errors) != 1 else "")
            return errors
        except subprocess.CalledProcessError as error:
            LOG.warning("Error calling pyre: %s", str(error))
            return []


def errors_from_run(_arguments) -> List[Dict[str, Any]]:
    configuration_path = Configuration.find_project_configuration()
    if not configuration_path:
        LOG.warning("Could not find pyre configuration.")
        return []
    with open(configuration_path) as configuration_file:
        configuration = Configuration(configuration_path, json.load(configuration_file))
        errors = configuration.get_errors()
        return filter_errors(_arguments, errors)


def remove_comment_preamble(lines: List[str]) -> None:
    # Deprecated: leaving remove logic until live old-style comments are cleaned up.
    while lines:
        old_line = lines.pop()
        new_line = re.sub(r"# pyre: .*$", "", old_line).rstrip()
        if old_line == "" or new_line != "":
            # The preamble has ended.
            lines.append(new_line)
            return


def _split_across_lines(
    comment: str, indent: int, max_line_length: Optional[int]
) -> List[str]:
    if not max_line_length or len(comment) <= max_line_length:
        return [comment]

    comment = comment.lstrip()
    available_columns = max_line_length - indent - len("#  ")

    buffered_line = ""
    result = []
    prefix = " " * indent
    for token in comment.split():
        if buffered_line and (
            len(buffered_line) + len(token) + len(" ") > available_columns
        ):
            # This new token would make the line exceed the limit,
            # hence terminate what we have accumulated.
            result.append(("{}{}".format(prefix, buffered_line)).rstrip())
            # The first line already has a comment token on it, so don't prefix #. For
            # the rest, we need to add the comment symbol manually.
            prefix = "{}#  ".format(" " * indent)
            buffered_line = ""

        buffered_line = buffered_line + token + " "

    result.append(("{}{}".format(prefix, buffered_line)).rstrip())
    return result


@verify_stable_ast
def fix_file(
    arguments: argparse.Namespace,
    filename: str,
    errors: Dict[int, List[Dict[str, str]]],
) -> None:
    custom_comment = arguments.comment if hasattr(arguments, "comment") else ""
    max_line_length = (
        arguments.max_line_length if arguments.max_line_length > 0 else None
    )
    path = Path(filename)
    text = path.read_text()
    if "@" "generated" in text:
        LOG.warning("Attempting to upgrade generated file %s, skipping.", filename)
        return
    lines = text.split("\n")  # type: List[str]

    # Replace lines in file.
    new_lines = []
    removing_pyre_comments = False
    for index, line in enumerate(lines):
        if removing_pyre_comments:
            # Only delete continuation comments of the form
            # "# pyre-fixme[2]:\n#  expected type `T`."
            if line.lstrip().startswith("#  "):
                continue
            else:
                removing_pyre_comments = False
        number = index + 1
        if number not in errors:
            new_lines.append(line)
            continue
        if errors[number][0]["code"] == "0":
            # Handle unused ignores.
            removing_pyre_comments = True
            replacement = re.sub(r"# pyre-(ignore|fixme).*$", "", line).rstrip()
            if replacement == "":
                remove_comment_preamble(new_lines)
            else:
                new_lines.append(replacement)
            continue

        comments = []
        for error in errors[number]:
            indent = len(line) - len(line.lstrip(" "))
            description = custom_comment if custom_comment else error["description"]
            comment = "{}# pyre-fixme[{}]: {}".format(
                " " * indent, error["code"], description
            )

            if not max_line_length or len(comment) <= max_line_length:
                comments.append(comment)
            else:
                if arguments.truncate:
                    comments.append(comment[: (max_line_length - 3)] + "...")
                else:
                    comments.extend(
                        _split_across_lines(comment, indent, max_line_length)
                    )

        LOG.info(
            "Adding comment%s on line %d: %s",
            "s" if len(comments) > 1 else "",
            number,
            " \n".join(comments),
        )
        new_lines.extend(comments)
        new_lines.append(line)
    new_text = "\n".join(new_lines)
    path.write_text(new_text)


# Exposed for testing.
def _upgrade_project(
    arguments: argparse.Namespace,
    configuration: Configuration,
    root: Path,
    version_control,
) -> None:
    LOG.info("Processing %s", configuration.get_directory())
    if not configuration.is_local or not configuration.version:
        return
    configuration.remove_version()
    errors = (
        errors_from_stdin(arguments)
        if arguments.from_stdin
        else configuration.get_errors()
    )
    if len(errors) > 0:
        fix(arguments, sort_errors(errors))

        # Lint and re-run pyre once to resolve most formatting issues
        if arguments.lint:
            lint_status = get_lint_status()
            if lint_status:
                apply_lint()
                errors = configuration.get_errors(should_clean=False)
                fix(arguments, sort_errors(errors))
    try:
        project_root = root.resolve()
        local_root = configuration.get_directory().resolve()
        version_control.submit_changes(
            arguments.submit,
            version_control.commit_message(str(local_root.relative_to(project_root))),
        )
    except subprocess.CalledProcessError:
        LOG.info("Error while running hg.")


def fix(arguments: argparse.Namespace, result: List[Tuple[str, List[Any]]]) -> None:
    for path, errors in result:
        LOG.info("Processing `%s`", path)

        # Build map from line to error codes.
        error_map = defaultdict(lambda: [])
        for error in errors:
            if error["concise_description"]:
                description = error["concise_description"]
            else:
                description = error["description"]
            match = re.search(r"\[(\d+)\]: (.*)", description)
            if match:
                error_map[error["line"]].append(
                    {"code": match.group(1), "description": match.group(2)}
                )
        fix_file(arguments, path, error_map)


@verify_stable_ast
def add_local_unsafe(arguments: argparse.Namespace, filename: str) -> None:
    LOG.info("Processing `%s`", filename)
    path = Path(filename)
    text = path.read_text()
    if "@" "generated" in text:
        LOG.warning("Attempting to edit generated file %s, skipping.", filename)
        return

    lines = text.split("\n")  # type: List[str]

    # Check if already locally strict or ignore-all.
    for line in lines:
        if re.match("^[ \t]*# *pyre-strict *$", line) or re.match(
            "^[ \t]*# *pyre-ignore-all-errors *$", line
        ):
            return

    def is_header(line: str) -> bool:
        is_comment = line.lstrip().startswith("#")
        is_pyre_ignore = (
            re.match("^[ \t]*# *pyre-ignore *$", line)
            or re.match("^[ \t]*# *pyre-fixme *$", line)
            or re.match("^[ \t]*# *type: ignore *$", line)
        )
        return is_comment and not is_pyre_ignore

    # Add local unsafe.
    new_lines = []
    past_header = False
    for line in lines:
        if not past_header and not is_header(line):
            past_header = True
            if len(new_lines) != 0:
                new_lines.append("")
            new_lines.append("# pyre-unsafe")
        new_lines.append(line)
    new_text = "\n".join(new_lines)
    path.write_text(new_text)


def run_global_version_update(arguments: argparse.Namespace, version_control) -> None:
    global_configuration = Configuration.find_project_configuration()
    if global_configuration is None:
        LOG.error("No global configuration file found.")
        return

    with open(global_configuration, "r") as global_configuration_file:
        configuration = json.load(global_configuration_file)
        if "version" not in configuration:
            LOG.error(
                "Global configuration at %s has no version field.", global_configuration
            )
            return

        old_version = configuration["version"]

    # Rewrite.
    with open(global_configuration, "w") as global_configuration_file:
        configuration["version"] = arguments.hash

        # This will sort the keys in the configuration - we won't be clobbering comments
        # since Python's JSON parser disallows comments either way.
        json.dump(configuration, global_configuration_file, sort_keys=True, indent=2)
        global_configuration_file.write("\n")

    paths = arguments.paths
    configuration_paths = (
        [path / ".pyre_configuration.local" for path in paths]
        if paths
        else [
            configuration.get_path()
            for configuration in Configuration.gather_local_configurations(arguments)
            if configuration.is_local
        ]
    )
    for configuration_path in configuration_paths:
        if "mock_repository" in str(configuration_path):
            # Skip local configurations we have for testing.
            continue
        with open(configuration_path) as configuration_file:
            contents = json.load(configuration_file)
            if "version" in contents:
                LOG.info(
                    "Skipping %s as it already has a custom version field.",
                    configuration_path,
                )
                continue
            contents["version"] = old_version

        with open(configuration_path, "w") as configuration_file:
            json.dump(contents, configuration_file, sort_keys=True, indent=2)
            configuration_file.write("\n")

    try:
        commit_summary = "Automatic upgrade to hash `{}`".format(arguments.hash)
        version_control.submit_changes(
            arguments.submit,
            version_control.commit_message(
                "global configuration", summary_override=commit_summary
            ),
        )
    except subprocess.CalledProcessError:
        LOG.info("Error while running hg.")


def run_strict_default(arguments: argparse.Namespace, _version_control) -> None:
    project_configuration = Configuration.find_project_configuration()
    if project_configuration is None:
        LOG.info("No project configuration found for the given directory.")
        return
    local_configuration = arguments.local_configuration
    if local_configuration:
        configuration_path = local_configuration / ".pyre_configuration.local"
    else:
        configuration_path = project_configuration
    with open(configuration_path) as configuration_file:
        configuration = Configuration(configuration_path, json.load(configuration_file))
        LOG.info("Processing %s", configuration.get_directory())
        configuration.add_strict()
        errors = configuration.get_errors()

        if len(errors) > 0:
            result = sort_errors(errors)
            for filename, _ in result:
                add_local_unsafe(arguments, filename)

            if arguments.lint:
                lint_status = get_lint_status()
                if lint_status:
                    apply_lint()


def run_fixme(arguments: argparse.Namespace, version_control) -> None:
    if arguments.run:
        errors = errors_from_run(arguments)
        fix(arguments, sort_errors(errors))

        if arguments.lint:
            lint_status = get_lint_status()
            if lint_status:
                apply_lint()
                errors = errors_from_run(arguments)
                fix(arguments, sort_errors(errors))
    else:
        errors = errors_from_stdin(arguments)
        fix(arguments, sort_errors(errors))


def run_fixme_single(arguments: argparse.Namespace, version_control) -> None:
    project_configuration = Configuration.find_project_configuration()
    if project_configuration is None:
        LOG.info("No project configuration found for the given directory.")
        return
    configuration_path = arguments.path / ".pyre_configuration.local"
    with open(configuration_path) as configuration_file:
        configuration = Configuration(configuration_path, json.load(configuration_file))
        _upgrade_project(
            arguments, configuration, project_configuration.parent, version_control
        )


def run_fixme_all(arguments: argparse.Namespace, version_control) -> None:
    # Create sandcastle command.
    if arguments.sandcastle:
        configurations = Configuration.gather_local_configurations(arguments)
        paths = [str(configuration.get_directory()) for configuration in configurations]
        with open(arguments.sandcastle) as sandcastle_file:
            sandcastle_command = json.load(sandcastle_file)
        if arguments.hash:
            sandcastle_command["args"]["hash"] = arguments.hash
        sandcastle_command["args"]["paths"] = paths
        sandcastle_command["args"]["push_blocking_only"] = arguments.push_blocking_only
        command = ["scutil", "create"]
        subprocess.run(command, input=json.dumps(sandcastle_command).encode())
        return

    # Run locally.
    if arguments.hash and isinstance(arguments.hash, str):
        run_global_version_update(arguments, version_control)

    project_configuration = Configuration.find_project_configuration()
    if project_configuration is None:
        LOG.info("No project configuration found for the current directory.")
        return

    configurations = Configuration.gather_local_configurations(arguments)
    for configuration in configurations:
        _upgrade_project(
            arguments, configuration, project_configuration.parent, version_control
        )


def run_fixme_targets_file(
    arguments: argparse.Namespace,
    project_directory: Path,
    path: str,
    target_names: List[str],
) -> None:
    LOG.info("Processing %s/TARGETS...", path)
    targets = [path + ":" + name + "-typecheck" for name in target_names]
    buck_test_command = ["buck", "test", "--show-full-json-output"] + targets

    def get_errors(path: str) -> Optional[List[Dict[str, Any]]]:
        buck_test = subprocess.run(
            buck_test_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if buck_test.returncode == 0:
            # Successful run with no type errors
            LOG.info("No errors in %s/TARGETS...", path)
        elif buck_test.returncode == 32:
            buck_test_output = buck_test.stdout.decode().split("\n")
            pyre_error_pattern = re.compile(
                r"\W*(.*\.pyi?):(\d*):(\d*) (.* \[(\d*)\]: .*)"
            )
            errors = {}
            for output_line in buck_test_output:
                matched = pyre_error_pattern.match(output_line)
                if matched:
                    path = matched.group(1)
                    line = int(matched.group(2))
                    column = int(matched.group(3))
                    description = matched.group(4)
                    code = matched.group(5)
                    error = {
                        "line": line,
                        "column": column,
                        "path": project_directory / path,
                        "code": code,
                        "description": description,
                        "concise_description": description,
                    }
                    errors[(line, column, path, code)] = error
            return list(errors.values())
        else:
            LOG.error(
                "Failed to run buck test command:\n\t%s\n\n%s",
                " ".join(buck_test_command),
                buck_test.stderr.decode(),
            )

    errors = get_errors(path)
    if not errors:
        return
    LOG.info("Found %d type errors in %s/TARGETS.", len(errors), path)
    if not errors:
        return
    fix(arguments, sort_errors(errors))
    if not arguments.lint:
        return
    lint_status = get_lint_status()
    if lint_status:
        LOG.info("Linting...")
        apply_lint()
        errors = get_errors(path)
        if not errors:
            LOG.info("Errors unchanged after linting.")
            return
        LOG.info("Found %d type errors after linting.", len(errors))
        fix(arguments, sort_errors(errors))


def run_fixme_targets(arguments: argparse.Namespace, version_control) -> None:
    # Currently does not support sandcastle integration, or setting the global hash
    # at the same time. As-is, run this locally after the global hash is updated.
    subdirectory = arguments.subdirectory
    subdirectory = Path(subdirectory) if subdirectory else None
    project_configuration = Configuration.find_project_configuration(subdirectory)
    if project_configuration is None:
        LOG.error("No project configuration found for the given directory.")
        return
    project_directory = project_configuration.parent
    search_root = subdirectory if subdirectory else project_directory
    LOG.info("Finding typecheck targets in %s", search_root)
    # TODO(T56778370): Clean up code by parsing the TARGETS file rather than using grep.
    typing_field = "check_types ?= ?True"
    typing_options_field = 'check_types_options ?= ?"[^"]*pyre[^"]*",?'
    targets_regex = r"(?s)name = ((?!\n\s*name).)*{}((?!\n\s*name).)*{}".format(
        typing_field, typing_options_field
    )
    find_targets_command = [
        "grep",
        "-RPzo",
        "--include=*TARGETS",
        targets_regex,
        search_root,
    ]
    find_targets = subprocess.run(
        find_targets_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    if find_targets.returncode == 1:
        LOG.info("Did not find any targets to upgrade.")
        return
    if find_targets.returncode != 0:
        LOG.error("Failed to search for targets: %s", find_targets.stderr.decode())
        return
    output = find_targets.stdout.decode()
    targets = re.split(typing_options_field, output)
    target_pattern = re.compile(r"(.*)\/TARGETS:.*name = \"([^\"]*)\".*")
    target_names = {}
    total_targets = 0
    for target in targets:
        matched = target_pattern.match(target.replace("\n", " ").strip())
        if matched:
            total_targets += 1
            path = matched.group(1).strip()
            target_name = matched.group(2)
            if path in target_names:
                target_names[path].append(target_name)
            else:
                target_names[path] = [target_name]
    LOG.info(
        "Found %d typecheck targets in %d TARGETS files to analyze",
        total_targets,
        len(target_names),
    )
    for path, target_names in target_names.items():
        run_fixme_targets_file(arguments, project_directory, path, target_names)
    try:
        if not arguments.no_commit:
            version_control.submit_changes(
                arguments.submit,
                version_control.commit_message("{} (TARGETS)".format(search_root)),
            )
    except subprocess.CalledProcessError:
        LOG.info("Error while running hg.")


def path_exists(filename: str) -> Path:
    path = Path(filename)
    if not path.exists():
        raise ValueError("No file at {}".format(filename))
    return path


def run(version_control) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", action="store_true", help="Enable verbose logging")
    parser.add_argument(
        "--truncate",
        action="store_true",
        help="Truncate error messages to maximum line length.",
    )
    parser.add_argument(
        "--max-line-length",
        default=88,
        type=int,
        help="Enforce maximum line length on new comments "
        + "(default: %(default)s, use 0 to set no maximum line length)",
    )
    parser.add_argument(
        "--only-fix-error-code",
        type=int,
        help="Only add fixmes for errors with this specific error code.",
        default=None,
    )

    commands = parser.add_subparsers()

    # Subcommands: Codemods
    missing_overridden_return_annotations = commands.add_parser(
        "missing-overridden-return-annotations",
        help="Add annotations according to errors inputted through stdin.",
    )
    missing_overridden_return_annotations.set_defaults(
        function=run_missing_overridden_return_annotations
    )

    missing_global_annotations = commands.add_parser(
        "missing-global-annotations",
        help="Add annotations according to errors inputted through stdin.",
    )
    missing_global_annotations.set_defaults(function=run_missing_global_annotations)

    # Subcommand: Change default pyre mode to strict and adjust module headers.
    strict_default = commands.add_parser("strict-default")
    strict_default.set_defaults(function=run_strict_default)
    strict_default.add_argument(
        "-l",
        "--local-configuration",
        type=path_exists,
        help="Path to project root with local configuration",
    )
    strict_default.add_argument(
        # TODO(T53195818): Not implemented
        "--remove-strict-headers",
        action="store_true",
        help="Delete unnecessary `# pyre-strict` headers.",
    )
    strict_default.add_argument("--lint", action="store_true", help=argparse.SUPPRESS)

    # Subcommand: Set global configuration to given hash, and add version override
    # to all local configurations to run previous version.
    update_global_version = commands.add_parser("update-global-version")
    update_global_version.set_defaults(function=run_global_version_update)
    update_global_version.add_argument("hash", help="Hash of new Pyre version")
    update_global_version.add_argument(
        "-p", "--push-blocking-only", action="store_true"
    )
    update_global_version.add_argument(
        "--paths",
        nargs="*",
        help="A list of paths to local Pyre projects.",
        default=[],
        type=path_exists,
    )
    update_global_version.add_argument(
        "--submit", action="store_true", help=argparse.SUPPRESS
    )

    # Subcommand: Fixme all errors inputted through stdin.
    fixme = commands.add_parser("fixme")
    fixme.set_defaults(function=run_fixme)
    fixme.add_argument("--comment", help="Custom comment after fixme comments")
    fixme.add_argument("--run", action="store_true")
    fixme.add_argument("--lint", action="store_true", help=argparse.SUPPRESS)

    # Subcommand: Fixme all errors for a single project.
    fixme_single = commands.add_parser("fixme-single")
    fixme_single.set_defaults(function=run_fixme_single)
    fixme_single.add_argument(
        "path", help="Path to project root with local configuration", type=path_exists
    )
    fixme_single.add_argument(
        "--from-stdin", action="store_true", help=argparse.SUPPRESS
    )
    fixme_single.add_argument("--submit", action="store_true", help=argparse.SUPPRESS)
    fixme_single.add_argument("--lint", action="store_true", help=argparse.SUPPRESS)

    # Subcommand: Fixme all errors in all projects with local configurations.
    fixme_all = commands.add_parser("fixme-all")
    fixme_all.set_defaults(function=run_fixme_all)
    fixme_all.add_argument(
        "-c", "--comment", help="Custom comment after fixme comments"
    )
    fixme_all.add_argument("-p", "--push-blocking-only", action="store_true")
    fixme_all.add_argument("--submit", action="store_true", help=argparse.SUPPRESS)
    fixme_all.add_argument("--lint", action="store_true", help=argparse.SUPPRESS)
    fixme_all.add_argument(
        "-s",
        "--sandcastle",
        help="Create upgrade stack on sandcastle.",
        type=path_exists,
    )
    fixme_all.add_argument(
        "hash", nargs="?", default=None, help="Hash of new Pyre version"
    )

    # Subcommand: Fixme all errors in targets running type checking
    fixme_targets = commands.add_parser("fixme-targets")
    fixme_targets.set_defaults(function=run_fixme_targets)
    fixme_targets.add_argument(
        "-c", "--comment", help="Custom comment after fixme comments"
    )
    fixme_targets.add_argument("--submit", action="store_true", help=argparse.SUPPRESS)
    fixme_targets.add_argument("--lint", action="store_true", help=argparse.SUPPRESS)
    fixme_targets.add_argument(
        "--subdirectory", help="Only upgrade TARGETS files within this directory."
    )
    fixme_targets.add_argument(
        "--no-commit", action="store_true", help="Keep changes in working state."
    )
    # Initialize default values.
    arguments = parser.parse_args()
    if not hasattr(arguments, "function"):
        # pyre-fixme[16]: `Namespace` has no attribute `run`.
        arguments.run = False
        # pyre-fixme[16]: `Namespace` has no attribute `function`.
        arguments.function = run_fixme

    # Initialize values that may be null-checked, but do not exist as a flag
    # for all subcommands
    if not hasattr(arguments, "paths"):
        # pyre-fixme[16]: `Namespace` has no attribute `paths`.
        arguments.paths = None
    if not hasattr(arguments, "from_stdin"):
        # pyre-fixme[16]: `Namespace` has no attribute `from_stdin`.
        arguments.from_stdin = None

    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.DEBUG if arguments.verbose else logging.INFO,
    )

    try:
        exit_code = ExitCode.SUCCESS
        arguments.function(arguments, version_control)
    except Exception as error:
        LOG.error(str(error))
        LOG.info(traceback.format_exc())
        exit_code = ExitCode.FAILURE

    sys.exit(exit_code)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import copy
import io
import logging
import os
import re
import sys
import threading
import time
from typing import List, Optional, Sequence  # noqa


PERFORMANCE: int = 15
PROMPT: int = 50
SUCCESS: int = 60


LOG = logging.getLogger(__name__)  # type: logging.Logger


stdout = io.StringIO(newline="")  # type: io.StringIO
__handler: Optional["TimedStreamHandler"] = None


class Color:
    YELLOW = "\033[33m"  # type: str
    RED = "\033[31m"  # type: str
    GREEN = "\033[32m"  # type: str


class Format:
    BOLD = "\033[1m"  # type: str

    CLEAR_LINE = "\x1b[0G\x1b[K"  # type: str
    CLEAR = "\033[0m"  # type: str
    TRUNCATE_OVERFLOW = "\033[?7l"  # type: str
    WRAP_OVERFLOW = "\033[?7h"  # type: str
    NEWLINE = "\n"  # type: str

    CURSOR_UP_LINE = "\x1b[1A"  # type: str
    HIDE_CURSOR = "\x1b[?25l"  # type: str
    SHOW_CURSOR = "\x1b[?25h"  # type: str


class Character:
    LAMBDA = "ƛ"  # type: str


class SectionFormatter(logging.Formatter):
    def __init__(self) -> None:
        super(SectionFormatter, self).__init__("%(asctime)s %(levelname)s %(message)s")

    def format(self, record: logging.LogRecord) -> str:
        formatted = super(SectionFormatter, self).format(record)
        return re.sub(r"DEBUG \[(.*)\]", r"\1", formatted)


class TimedStreamHandler(logging.StreamHandler):
    THRESHOLD = 0.5  # type: float
    LINE_BREAKING_LEVELS = ["ERROR", "WARNING", "SUCCESS"]  # type: Sequence[str]

    _terminate = False  # type: bool
    _last_update = 0.0  # type: float

    def __init__(self) -> None:
        super(TimedStreamHandler, self).__init__()
        self.setFormatter(logging.Formatter("%(message)s"))
        self.terminator = ""  # type: str
        self.setLevel(logging.INFO)

        self._record = None  # type: Optional[logging.LogRecord]
        self._last_record = None  # type: Optional[logging.LogRecord]
        self._active_lines = 0  # type: int

        # Preamble preparing terminal.
        sys.stderr.write(
            Format.NEWLINE
            + Format.CLEAR_LINE
            + Format.CURSOR_UP_LINE
            + Format.HIDE_CURSOR
        )

        thread = threading.Thread(target=self._thread)
        thread.daemon = True
        thread.start()

    def clear_lines(self) -> str:
        if self._active_lines == 0:
            return ""
        return Format.CLEAR_LINE + "".join(
            [
                Format.CURSOR_UP_LINE + Format.CLEAR_LINE
                for n in range(self._active_lines - 1)
            ]
        )

    def emit(self, record: logging.LogRecord, age: Optional[float] = None) -> None:
        self._last_record = record
        suffix = ""
        color = ""
        active_lines = record.msg.count("\n") + 1
        if record.levelname in self.LINE_BREAKING_LEVELS:
            record.msg += "\n"

        if record.levelname == "ERROR":
            color = Color.RED
            self._record = None
            active_lines = 0
        elif record.levelname == "WARNING":
            color = Color.YELLOW
            self._record = None
            active_lines = 0
        elif record.levelname == "PROMPT":
            color = Color.YELLOW
            self._record = None
            active_lines = 0
        elif record.levelname == "SUCCESS":
            self._record = None
            active_lines = 0
        elif age:
            if age > 10:
                color = Color.YELLOW
            if age > 30:
                color = Color.RED
            suffix = " {}[{:.1f}s]{}".format(
                color if color else "", age, Format.CLEAR if color else ""
            )
        else:
            self._record = record
            self._last_update = time.time()

        timed_record = copy.copy(record)
        timed_record.msg = (
            "{clear_line}{color} {cursor}{clear} " "{truncate}{message}{suffix}"
        ).format(
            clear_line=self.clear_lines(),
            color=color,
            cursor=Character.LAMBDA,
            clear=Format.CLEAR,
            truncate=Format.TRUNCATE_OVERFLOW,
            message=record.msg,
            suffix=suffix,
        )
        self._active_lines = active_lines
        super(TimedStreamHandler, self).emit(timed_record)

    def _thread(self) -> None:
        while not self._terminate:
            record = self._record
            if record:
                age = time.time() - self._last_update
                if age > self.THRESHOLD:
                    self.emit(record, age)
            time.sleep(0.1)

    def terminate(self) -> None:
        last_record = self._last_record
        if last_record and last_record.levelname not in self.LINE_BREAKING_LEVELS:
            sys.stderr.write("\n")

        # Reset terminal.
        sys.stderr.write(Format.WRAP_OVERFLOW + Format.SHOW_CURSOR)
        sys.stderr.flush()
        self._terminate = True


def initialize(
    noninteractive: bool,
    log_directory: str = "/tmp/.pyre",
    disable_file_logging: bool = False,
) -> None:
    global __handler
    if noninteractive:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(SectionFormatter())
        stream_handler.setLevel(logging.DEBUG)
        __handler = None
    else:
        stream_handler = TimedStreamHandler()
        __handler = stream_handler

    handlers = [stream_handler]  # type: List[logging.Handler]

    if not noninteractive and not disable_file_logging:
        if not os.path.exists(log_directory):
            os.makedirs(log_directory)
        file_handler = logging.FileHandler(os.path.join(log_directory, "pyre.stderr"))
        file_handler.setFormatter(SectionFormatter())
        file_handler.setLevel(logging.DEBUG)
        handlers.append(file_handler)
    logging.addLevelName(PERFORMANCE, "PERFORMANCE")
    logging.addLevelName(PROMPT, "PROMPT")
    logging.addLevelName(SUCCESS, "SUCCESS")
    logging.basicConfig(level=logging.DEBUG, handlers=handlers)


def cleanup() -> None:
    handler = __handler
    if handler:
        handler.terminate()

    output = stdout.getvalue()
    if output:
        sys.stdout.write(output + "\n")


class Buffer:
    THRESHOLD = 0.1  # type: float

    _flushed = False  # type: bool

    def __init__(self, section: str, data: List[str]) -> None:
        self._section = section  # type: str
        self._data = data  # type: List[str]
        self._lock = threading.RLock()  # type: threading.RLock
        thread = threading.Thread(target=self._thread)
        thread.daemon = True
        thread.start()

    def append(self, line: str) -> None:
        self._data.append(line)

    def flush(self) -> None:
        with self._lock:
            if self._flushed is True:
                return
            self._flushed = True
        message = "\n".join(self._data)
        if self._section == "ERROR":
            LOG.error(message)
        elif self._section == "INFO":
            LOG.info(message)
        elif self._section == "DUMP":
            LOG.warning(message)
        elif self._section == "WARNING":
            LOG.warning(message)
        elif self._section == "PROGRESS":
            LOG.info(message)
        elif self._section == "PARSER":
            LOG.error(message)
        else:
            LOG.debug("[%s] %s", self._section, message)

    def _thread(self) -> None:
        time.sleep(self.THRESHOLD)
        with self._lock:
            if not self._flushed:
                self.flush()


def get_yes_no_input(prompt: str) -> bool:
    choice = get_input(prompt, suffix=" [Y/n] ")
    return choice.lower() in ["", "y", "ye", "yes"]


def get_optional_input(prompt: str, default: str) -> str:
    result = get_input(prompt, suffix=" (Default: `{}`): ".format(default))
    if result == "":
        return default
    return result


def get_input(prompt: str, suffix: str = "") -> str:
    LOG.log(PROMPT, prompt + suffix)
    return input().strip()



# Copyright (c) 2019-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import logging
import os
import socket
from types import TracebackType
from typing import BinaryIO, Optional

from . import json_rpc


LOG: logging.Logger = logging.getLogger(__name__)


class SocketException(Exception):
    pass


class SocketConnection(object):
    def __init__(self, root: str) -> None:
        self.socket = socket.socket(
            socket.AF_UNIX, socket.SOCK_STREAM
        )  # type: socket.socket
        self.root = root
        self.input: BinaryIO = self.socket.makefile(mode="rb")
        self.output: BinaryIO = self.socket.makefile(mode="wb")

    def connect(self) -> "SocketConnection":
        socket_path = self._socket_path()
        try:
            self.socket.connect(os.path.realpath(socket_path))
            return self
        except (ConnectionRefusedError, FileNotFoundError, OSError) as error:
            raise SocketException(
                "Failed to connect to server at `{}`. Reason: `{}`".format(
                    socket_path, error
                )
            )

    def perform_handshake(self, version_hash: str) -> None:
        try:
            json_rpc.perform_handshake(self.input, self.output, version_hash)
        except (OSError, ValueError) as error:
            raise SocketException(
                "Exception encountered during handshake: `{}`".format(error)
            )

    def send_request(self, request: json_rpc.Request) -> None:
        if not request.write(self.output):
            LOG.info("Failed to communicate with server. Shutting down.")
            raise SocketException

    def _socket_path(self) -> str:
        return os.path.join(self.root, "server", "json_server.sock")

    def __enter__(self) -> "SocketConnection":
        self.connect()
        return self

    def __exit__(
        self,
        _type: Optional[BaseException],
        _value: Optional[BaseException],
        _traceback: Optional[TracebackType],
    ) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()

    def close(self) -> None:
        try:
            self.socket.close()
        except OSError:
            pass
